import sys

import pytest

pytest.importorskip("jax")

import jax
from jax import numpy as jnp
from numpy import ndarray

import nifty8.re as jft

NDARRAY_TYPE = [ndarray]

try:
    from jax.numpy import ndarray as jndarray

    NDARRAY_TYPE.append(jndarray)
except ImportError:
    pass

NDARRAY_TYPE = tuple(NDARRAY_TYPE)


def _json_serialize(obj):
    if isinstance(obj, NDARRAY_TYPE):
        return obj.tolist()
    raise TypeError(f"unknown type {type(obj)}")


def hashit(obj, n_chars=8) -> str:
    """Get first `n_chars` characters of Blake2B hash of `obj`."""
    import hashlib
    import json

    return hashlib.blake2b(
        bytes(json.dumps(obj, default=_json_serialize), "utf-8")
    ).hexdigest()[:n_chars]


def test_hmc_hash():
    """Test sapmler output against known hash from previous commits."""
    x0 = jnp.array([0.1, 1.223], dtype=jnp.float32)
    sampler = jft.HMCChain(
        potential_energy=lambda x: jnp.sum(x**2),
        inverse_mass_matrix=1.,
        position_proto=x0,
        step_size=0.193,
        num_steps=100,
        max_energy_difference=1.
    )
    chain, (key, pos) = sampler.generate_n_samples(
        key=42, initial_position=x0, num_samples=1000, save_intermediates=True
    )
    assert chain.divergences.sum() == 0
    accepted = chain.trees.accepted
    results = (pos, key, chain.samples, accepted)
    results_hash = hashit(results, n_chars=20)
    print(f"full hash: {results_hash}", file=sys.stderr)
    old_hash = "3d665689f809a98c81b3"
    assert results_hash == old_hash


def test_nuts_hash():
    """Test sapmler output against known hash from previous commits."""
    jax.config.update("jax_enable_x64", False)

    x0 = jnp.array([0.1, 1.223], dtype=jnp.float32)
    sampler = jft.NUTSChain(
        potential_energy=lambda x: jnp.sum(x**2),
        inverse_mass_matrix=1.,
        position_proto=x0,
        step_size=0.193,
        max_tree_depth=10,
        bias_transition=False,
        max_energy_difference=1.
    )
    chain, (key, pos) = sampler.generate_n_samples(
        key=42, initial_position=x0, num_samples=1000, save_intermediates=False
    )
    assert chain.divergences.sum() == 0
    results = (pos, key, chain.samples)
    results_hash = hashit(results, n_chars=20)
    print(f"full hash: {results_hash}", file=sys.stderr)
    old_hash = "8043850d7249acb77b26"
    assert results_hash == old_hash

    jax.config.update("jax_enable_x64", True)


if __name__ == "__main__":
    test_hmc_hash()
    test_nuts_hash()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from ..utilities import check_dtype_or_none
from .endomorphic_operator import EndomorphicOperator


class ScalingOperator(EndomorphicOperator):
    """Operator which multiplies a Field with a scalar.

    Parameters
    ----------
    domain : Domain or tuple of Domain or DomainTuple
        The domain on which the Operator's input Field is defined.
    factor : scalar
        The multiplication factor
    sampling_dtype :
        If this operator represents the covariance of a Gaussian probabilty
        distribution, `sampling_dtype` specifies if it is real or complex
        Gaussian. If `sampling_dtype` is `None`, the operator cannot be used as
        a covariance, i.e. no samples can be drawn. Default: None.

    Notes
    -----
    :class:`Operator` supports the multiplication with a scalar. So one does
    not need instantiate :class:`ScalingOperator` explicitly in most cases.

    Formally, this operator always supports all operation modes (times,
    adjoint_times, inverse_times and inverse_adjoint_times), even if `factor`
    is 0 or infinity. It is the user's responsibility to apply the operator
    only in appropriate ways (e.g. call inverse_times only if `factor` is
    nonzero).

    Along with this behaviour comes the feature that it is possible to draw an
    inverse sample from a :class:`ScalingOperator` (which is a zero-field).
    This occurs if one draws an inverse sample of a positive definite sum of
    two operators each of which are only positive semi-definite. However, it
    is unclear whether this beviour does not lead to unwanted effects
    somewhere else.
    """

    def __init__(self, domain, factor, sampling_dtype=None):
        from ..sugar import makeDomain

        if not np.isscalar(factor):
            raise TypeError("Scalar required")
        self._domain = makeDomain(domain)
        self._factor = factor
        self._capability = self._all_ops
        check_dtype_or_none(sampling_dtype, self._domain)
        self._dtype = sampling_dtype

        try:
            from functools import partial

            from jax import numpy as jnp

            self._jax_expr = partial(jnp.multiply, factor)
        except ImportError:
            self._jax_expr = None

    def apply(self, x, mode):
        from ..sugar import full

        self._check_input(x, mode)
        fct = self._factor
        if fct == 1.:
            return x
        if fct == 0.:
            return full(x.domain, 0.)

        MODES_WITH_ADJOINT = self.ADJOINT_TIMES | self.ADJOINT_INVERSE_TIMES
        MODES_WITH_INVERSE = self.INVERSE_TIMES | self.ADJOINT_INVERSE_TIMES
        if (mode & MODES_WITH_ADJOINT) != 0:
            fct = np.conj(fct)
        if (mode & MODES_WITH_INVERSE) != 0:
            fct = 1./fct
        return x*fct

    def _flip_modes(self, trafo):
        fct = self._factor
        if trafo & self.ADJOINT_BIT:
            fct = np.conj(fct)
        if trafo & self.INVERSE_BIT:
            fct = 1./fct
        return ScalingOperator(self._domain, fct, self._dtype)

    def _get_fct(self, from_inverse):
        fct = self._factor
        if (fct.imag != 0. or fct.real < 0. or
                (fct.real == 0. and from_inverse)):
            raise ValueError("operator not positive definite")
        return 1./np.sqrt(fct) if from_inverse else np.sqrt(fct)

    def draw_sample(self, from_inverse=False):
        from ..sugar import from_random
        if self._dtype is None:
            s = "Need to specify dtype to be able to sample from this operator:\n"
            s += self.__repr__()
            raise RuntimeError(s)
        return from_random(domain=self._domain, random_type="normal",
                           dtype=self._dtype, std=self._get_fct(from_inverse))

    def get_sqrt(self):
        fct = self._get_fct(False)
        if np.iscomplexobj(fct) or fct < 0:
            raise ValueError("get_sqrt() works only for positive definite operators.")
        return ScalingOperator(self._domain, fct)

    def __call__(self, other):
        res = EndomorphicOperator.__call__(self, other)
        if np.isreal(self._factor) and self._factor >= 0:
            if other.jac is not None and other.metric is not None:
                from .sandwich_operator import SandwichOperator
                sqrt_fac = np.sqrt(self._factor)
                newop = ScalingOperator(other.metric.domain, sqrt_fac, self._dtype)
                met = SandwichOperator.make(newop, other.metric)
                res = res.add_metric(met)
        return res

    def __repr__(self):
        s = f"ScalingOperator ({self._factor}"
        if self._dtype is not None:
            s += f", sampling dtype {self._dtype}"
        s += ")"
        return s



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from operator import add, sub

import numpy as np

from ..field import Field
from ..multi_field import MultiField
from ..sugar import makeDomain
from .operator import Operator


class Adder(Operator):
    """Adds a fixed field.

    Parameters
    ----------
    a : :class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField` or Scalar
        The field by which the input is shifted.
    """
    def __init__(self, a, neg=False, domain=None):
        self._a = a
        if isinstance(a, (Field, MultiField)):
            dom = a.domain
        elif np.isscalar(a):
            dom = makeDomain(domain)
        else:
            raise TypeError
        self._domain = self._target = dom
        self._neg = bool(neg)

        try:
            from jax.tree_util import tree_map

            from ..re import Vector as ReField

            a_j = ReField(a.val) if isinstance(a, (Field, MultiField)) else a

            def jax_expr(x):
                # Preserve the input type
                if not isinstance(x, ReField):
                    a_astype_x = a_j.tree if isinstance(a_j, ReField) else a_j
                else:
                    a_astype_x = a_j
                return tree_map(sub if neg else add, x, a_astype_x)

            self._jax_expr = jax_expr
        except ImportError:
            self._jax_expr = None

    def apply(self, x):
        self._check_input(x)
        if self._neg:
            return x - self._a
        return x + self._a



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest

from ..common import list2fixture, setup_function, teardown_function

_h_RG_spaces = [ift.RGSpace(7, distances=0.2, harmonic=True),
                ift.RGSpace((12, 46), distances=(.2, .3), harmonic=True)]
_h_spaces = _h_RG_spaces + [ift.LMSpace(17)]
_p_RG_spaces = [ift.RGSpace(19, distances=0.7),
                ift.RGSpace((1, 2, 3, 6), distances=(0.2, 0.25, 0.34, .8))]
_p_spaces = _p_RG_spaces + [ift.HPSpace(17), ift.GLSpace(8, 13)]
_pow_spaces = [ift.PowerSpace(ift.RGSpace((17, 38), (0.99, 1340), harmonic=True)),
               ift.PowerSpace(ift.LMSpace(18), ift.PowerSpace.useful_binbounds(ift.LMSpace(18), False))]

pmp = pytest.mark.parametrize
dtype = list2fixture([np.float64, np.complex128])


@pmp('sp', _p_RG_spaces)
def testLOSResponse(sp, dtype):
    starts = ift.random.current_rng().standard_normal((len(sp.shape), 10))
    ends = ift.random.current_rng().standard_normal((len(sp.shape), 10))
    sigma_low = 1e-4*ift.random.current_rng().standard_normal(10)
    sigma_ups = 1e-5*ift.random.current_rng().standard_normal(10)
    op = ift.LOSResponse(sp, starts, ends, sigma_low, sigma_ups)
    ift.extra.check_linear_operator(op, dtype, dtype)


@pmp('sp', _h_spaces + _p_spaces + _pow_spaces)
def testOperatorCombinations(sp, dtype):
    a = ift.DiagonalOperator(ift.Field.from_random(sp, "normal", dtype=dtype))
    b = ift.DiagonalOperator(ift.Field.from_random(sp, "normal", dtype=dtype))
    op = ift.SandwichOperator.make(a, b)
    ift.extra.check_linear_operator(op, dtype, dtype)
    op = a(b)
    ift.extra.check_linear_operator(op, dtype, dtype)
    op = a + b
    ift.extra.check_linear_operator(op, dtype, dtype)
    op = a - b
    ift.extra.check_linear_operator(op, dtype, dtype)


def testLinearInterpolator():
    sp = ift.RGSpace((10, 8), distances=(0.1, 3.5))
    pos = ift.random.current_rng().random((2, 23))
    pos[0, :] *= 0.9
    pos[1, :] *= 7*3.5
    op = ift.LinearInterpolator(sp, pos)
    ift.extra.check_linear_operator(op)


@pmp('sp', _h_spaces + _p_spaces + _pow_spaces)
def testRealizer(sp):
    op = ift.Realizer(sp)
    ift.extra.check_linear_operator(op, np.complex128, np.float64,
                                only_r_linear=True)


@pmp('sp', _h_spaces + _p_spaces + _pow_spaces)
def testImaginizer(sp):
    op = ift.Imaginizer(sp)
    ift.extra.check_linear_operator(op, np.complex128, np.float64,
                                only_r_linear=True)
    loc = ift.from_random(op.domain, dtype=np.complex128)
    ift.extra.check_operator(op, loc)


@pmp('sp', _h_spaces + _p_spaces + _pow_spaces)
def testConjugationOperator(sp):
    op = ift.ConjugationOperator(sp)
    ift.extra.check_linear_operator(op, np.complex128, np.complex128,
                                only_r_linear=True)


@pmp('sp', _h_spaces + _p_spaces + _pow_spaces)
def testOperatorAdaptor(sp, dtype):
    op = ift.DiagonalOperator(ift.Field.from_random(sp, "normal", dtype=dtype))
    ift.extra.check_linear_operator(op.adjoint, dtype, dtype)
    ift.extra.check_linear_operator(op.inverse, dtype, dtype)
    ift.extra.check_linear_operator(op.inverse.adjoint, dtype, dtype)
    ift.extra.check_linear_operator(op.adjoint.inverse, dtype, dtype)


@pmp('sp1', _h_spaces + _p_spaces + _pow_spaces)
@pmp('sp2', _h_spaces + _p_spaces + _pow_spaces)
def testNullOperator(sp1, sp2, dtype):
    op = ift.NullOperator(sp1, sp2)
    ift.extra.check_linear_operator(op, dtype, dtype)
    mdom1 = ift.MultiDomain.make({'a': sp1})
    mdom2 = ift.MultiDomain.make({'b': sp2})
    op = ift.NullOperator(mdom1, mdom2)
    ift.extra.check_linear_operator(op, dtype, dtype)
    op = ift.NullOperator(sp1, mdom2)
    ift.extra.check_linear_operator(op, dtype, dtype)
    op = ift.NullOperator(mdom1, sp2)
    ift.extra.check_linear_operator(op, dtype, dtype)


@pmp('sp', _p_RG_spaces)
def testHarmonicSmoothingOperator(sp, dtype):
    op = ift.HarmonicSmoothingOperator(sp, 0.1)
    ift.extra.check_linear_operator(op, dtype, dtype)


@pmp('sp', _h_spaces + _p_spaces + _pow_spaces)
def testDOFDistributor(sp, dtype):
    # TODO: Test for DomainTuple
    if sp.size < 4:
        return
    dofdex = np.arange(sp.size).reshape(sp.shape) % 3
    dofdex = ift.Field.from_raw(sp, dofdex)
    op = ift.DOFDistributor(dofdex)
    ift.extra.check_linear_operator(op, dtype, dtype)


@pmp('sp', _h_spaces)
def testPPO(sp, dtype):
    op = ift.PowerDistributor(target=sp)
    ift.extra.check_linear_operator(op, dtype, dtype)
    ps = ift.PowerSpace(
        sp, ift.PowerSpace.useful_binbounds(sp, logarithmic=False, nbin=3))
    op = ift.PowerDistributor(target=sp, power_space=ps)
    ift.extra.check_linear_operator(op, dtype, dtype)
    ps = ift.PowerSpace(
        sp, ift.PowerSpace.useful_binbounds(sp, logarithmic=True, nbin=3))
    op = ift.PowerDistributor(target=sp, power_space=ps)
    ift.extra.check_linear_operator(op, dtype, dtype)


@pmp('sp', _h_RG_spaces + _p_RG_spaces)
def testFFT(sp, dtype):
    op = ift.FFTOperator(sp)
    ift.extra.check_linear_operator(op, dtype, dtype)
    op = ift.FFTOperator(sp.get_default_codomain())
    ift.extra.check_linear_operator(op, dtype, dtype)


@pmp('sp', _h_RG_spaces + _p_RG_spaces)
def testHartley(sp, dtype):
    op = ift.HartleyOperator(sp)
    ift.extra.check_linear_operator(op, dtype, dtype)
    op = ift.HartleyOperator(sp.get_default_codomain())
    ift.extra.check_linear_operator(op, dtype, dtype)


@pmp('sp', _h_spaces)
def testHarmonic(sp, dtype):
    op = ift.HarmonicTransformOperator(sp)
    ift.extra.check_linear_operator(op, dtype, dtype)


@pmp('sp', _p_spaces)
def testMask(sp, dtype):
    f = ift.from_random(sp).val
    mask = np.zeros_like(f)
    mask[f > 0] = 1
    mask = ift.Field.from_raw(sp, mask)
    op = ift.MaskOperator(mask)
    ift.extra.check_linear_operator(op, dtype, dtype)


@pmp('sp', _h_spaces + _p_spaces)
def testDiagonal(sp, dtype):
    op = ift.DiagonalOperator(ift.Field.from_random(sp, dtype=dtype))
    ift.extra.check_linear_operator(op, dtype, dtype)


@pmp('sp', _h_spaces + _p_spaces + _pow_spaces)
def testGeometryRemover(sp, dtype):
    op = ift.GeometryRemover(sp)
    ift.extra.check_linear_operator(op, dtype, dtype)

@pmp('sp', _h_spaces + _p_spaces + _pow_spaces)
def testExtractAtIndices(sp, dtype):
    min_ax = np.min(sp.shape)
    n_ax = len(sp.shape)
    inds = (list(range(min_ax//2))+list(range(min_ax//3)), )*n_ax
    op = ift.ExtractAtIndices(sp, inds)
    ift.extra.check_linear_operator(op, dtype, dtype)


@pmp('spaces', [0, 1, 2, 3, (0, 1), (0, 2), (0, 1, 2), (0, 2, 3), (1, 3)])
@pmp('wgt', [0, 1, 2, -1])
def testContractionOperator(spaces, wgt, dtype):
    dom = (ift.RGSpace(1), ift.RGSpace(2), ift.GLSpace(3), ift.HPSpace(2))
    op = ift.ContractionOperator(dom, spaces, wgt)
    ift.extra.check_linear_operator(op, dtype, dtype)


def testDomainTupleFieldInserter():
    target = ift.DomainTuple.make((ift.UnstructuredDomain([3, 2]),
                                   ift.UnstructuredDomain(7),
                                   ift.RGSpace([4, 22])))
    op = ift.DomainTupleFieldInserter(target, 1, (5,))
    ift.extra.check_linear_operator(op)


@pmp('space', [0, 2])
@pmp('factor', [1, 2, 2.7])
@pmp('central', [False, True])
def testZeroPadder(space, factor, dtype, central):
    dom = (ift.RGSpace(4), ift.UnstructuredDomain(5), ift.RGSpace(3, 4),
           ift.HPSpace(2))
    newshape = [int(factor*ll) for ll in dom[space].shape]
    op = ift.FieldZeroPadder(dom, newshape, space, central)
    ift.extra.check_linear_operator(op, dtype, dtype)


@pmp('args', [[ift.RGSpace((13, 52, 40)), (4, 6, 25), None],
              [ift.RGSpace((128, 128)), (45, 48), 0],
              [ift.RGSpace(13), (7,), None],
              [(ift.HPSpace(3), ift.RGSpace((12, 24), distances=0.3)), (12, 12), 1]])
def testRegridding(args):
    op = ift.RegriddingOperator(*args)
    ift.extra.check_linear_operator(op)


@pmp('fdomain', [(ift.RGSpace((2, 3, 2)), ift.RGSpace((4,), distances=(7.,))),
                 ift.HPSpace(3)])
@pmp('domain', [(ift.RGSpace(2), ift.GLSpace(10)),
                ift.RGSpace((4, 3), distances=(0.1, 1.))])
def testOuter(fdomain, domain):
    f = ift.from_random(ift.makeDomain(fdomain), 'normal')
    op = ift.OuterProduct(domain, f)
    ift.extra.check_linear_operator(op)


@pmp('sp', _h_spaces + _p_spaces + _pow_spaces)
@pmp('seed', [12, 3])
def testValueInserter(sp, seed):
    with ift.random.Context(seed):
        ind = []
        for ss in sp.shape:
            if ss == 1:
                ind.append(0)
            else:
                ind.append(int(ift.random.current_rng().integers(0, ss-1)))
        op = ift.ValueInserter(sp, ind)
        ift.extra.check_linear_operator(op)


@pmp('sp', _pow_spaces)
def testSlopeRemover(sp):
    op = ift.library.correlated_fields._SlopeRemover(sp)
    ift.extra.check_linear_operator(op)


@pmp('sp', _pow_spaces)
def testTwoLogIntegrations(sp):
    op = ift.library.correlated_fields._TwoLogIntegrations(sp)
    ift.extra.check_linear_operator(op)


@pmp('sp', _h_spaces + _p_spaces + _pow_spaces)
def testSpecialSum(sp):
    op = ift.library.correlated_fields._SpecialSum(sp)
    ift.extra.check_linear_operator(op)


@pmp('dofdex', [(0,), (1,), (0, 1), (1, 0)])
def testCorFldDistr(dofdex):
    tgt = ift.UnstructuredDomain(len(dofdex))
    dom = ift.UnstructuredDomain(2)
    op = ift.library.correlated_fields._Distributor(dofdex, dom, tgt)
    ift.extra.check_linear_operator(op)


def metatestMatrixProductOperator(sp, mat_shape, seed, **kwargs):
    with ift.random.Context(seed):
        mat = ift.random.current_rng().standard_normal(mat_shape)
        op = ift.MatrixProductOperator(sp, mat, **kwargs)
        ift.extra.check_linear_operator(op)
        mat = mat + 1j*ift.random.current_rng().standard_normal(mat_shape)
        op = ift.MatrixProductOperator(sp, mat, **kwargs)
        ift.extra.check_linear_operator(op)


@pmp('sp', [ift.RGSpace(10)])
@pmp('spaces', [None, (0,)])
@pmp('seed', [12, 3])
def testMatrixProductOperator_1d(sp, spaces, seed):
    mat_shape = sp.shape * 2
    metatestMatrixProductOperator(sp, mat_shape, seed, spaces=spaces)


@pmp('sp', [ift.DomainTuple.make((ift.RGSpace((2)), ift.RGSpace((10))))])
@pmp('spaces', [(0,), (1,), (0, 1)])
@pmp('seed', [12, 3])
def testMatrixProductOperator_2d_spaces(sp, spaces, seed):
    appl_shape = []
    for sp_idx in spaces:
        appl_shape += sp[sp_idx].shape
    appl_shape = tuple(appl_shape)
    mat_shape = appl_shape * 2
    metatestMatrixProductOperator(sp, mat_shape, seed, spaces=spaces)


@pmp('sp', [ift.RGSpace((2, 10))])
@pmp('seed', [12, 3])
def testMatrixProductOperator_2d_flatten(sp, seed):
    appl_shape = (ift.utilities.my_product(sp.shape),)
    mat_shape = appl_shape * 2
    metatestMatrixProductOperator(sp, mat_shape, seed, flatten=True)


@pmp('seed', [12, 3])
def testPartialExtractor(seed):
    with ift.random.Context(seed):
        tgt = {'a': ift.RGSpace(1), 'b': ift.RGSpace(2)}
        dom = tgt.copy()
        dom['c'] = ift.RGSpace(3)
        dom = ift.MultiDomain.make(dom)
        tgt = ift.MultiDomain.make(tgt)
        op = ift.PartialExtractor(dom, tgt)
        ift.extra.check_linear_operator(op)


@pmp('seed', [12, 3])
def testSlowFieldAdapter(seed):
    dom = {'a': ift.RGSpace(1), 'b': ift.RGSpace(2)}
    op = ift.operators.simple_linear_operators._SlowFieldAdapter(dom, 'a')
    ift.extra.check_linear_operator(op)

@pmp('seed', [12, 3])
def testDiagonalExtractor(seed):
    N = 42
    square_space = ift.RGSpace([N,N])
    op = ift.library.variational_models.DiagonalSelector(square_space)
    ift.extra.check_linear_operator(op)

@pmp('seed', [12, 3])
@pmp("N", [10, 17])
def testLowerTriangularInserter(seed, N):
    square_space = ift.RGSpace([N, N])
    op = ift.library.variational_models.LowerTriangularInserter(square_space)
    ift.extra.check_linear_operator(op)

@pmp('seed', [12, 3])
def test_Multifield2Vector(seed):
    dom = {'a': ift.RGSpace(1), 'b': ift.RGSpace(2)}
    dom = ift.MultiDomain.make(dom)
    op = ift.Multifield2Vector(dom)
    ift.extra.check_linear_operator(op)



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import jax
import numpy as np
import pytest
from jax import random
from jax.scipy.ndimage import map_coordinates

from nifty8.re import SamplingCartesianGridLOS

jax.config.update("jax_enable_x64", True)

pmp = pytest.mark.parametrize


@pmp("seed", (42,))
@pmp("shape", ((10,), (25,), (12, 12), (6, 6, 6)))
def test_sampling_cartesian_grid_los(seed, shape):
    key = random.PRNGKey(seed)
    n_test_points = 1_000
    test_los = SamplingCartesianGridLOS(
        np.zeros((len(shape),)),
        np.ones((len(shape),))[np.newaxis],
        distances=tuple(1.0 / s for s in shape),
        shape=shape,
        n_sampling_points=n_test_points,
    )
    key, sk = random.split(key)
    for test_x in (np.ones(shape), random.normal(sk, shape)):
        desired = map_coordinates(
            test_x,
            [
                np.linspace(0, s - 1, num=n_test_points, endpoint=False)
                for s in test_x.shape
            ],
            order=1,
            cval=np.nan,
        ).mean()
        desired *= np.linalg.norm(np.ones((len(shape),)))
        np.testing.assert_allclose(
            test_los(test_x).sum(),
            desired,
            rtol=1e-2,
        )



import pytest

import nifty8.re as jft


def test_map_forest_axes_validation():
    f = lambda x: x
    jft.map_forest(f, in_axes=1)
    with pytest.raises(ValueError):
        jft.map_forest(f, in_axes=(None, ))
    with pytest.raises(TypeError):
        jft.map_forest(f, in_axes=(1., ))



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from .. import config
from . import structured_kernel_interpolation
from .conjugate_gradient import cg, static_cg
from .correlated_field import CorrelatedFieldMaker, non_parametric_amplitude
from .custom_map import lmap, smap
from .evi import (
    Samples, draw_linear_residual, draw_residual, nonlinearly_update_residual
)
from .evidence_lower_bound import estimate_evidence_lower_bound
from .extra import SamplingCartesianGridLOS
from .gauss_markov import (
    GaussMarkovProcess, IntegratedWienerProcess, OrnsteinUhlenbeckProcess,
    WienerProcess
)
from .hmc import generate_hmc_acc_rej, generate_nuts_tree
from .hmc_oo import HMCChain, NUTSChain
from .likelihood import Likelihood, LikelihoodPartial
from .likelihood_impl import (
    Categorical, Gaussian, Poissonian, StudentT, VariableCovarianceGaussian,
    VariableCovarianceStudentT
)
from .logger import logger
from .minisanity import minisanity, reduced_residual_stats
from .misc import hvp, interpolate, wrap, wrap_left
from .model import Initializer, Model, VModel, WrappedCall
from .num import *
from .optimize import minimize, newton_cg, trust_ncg
from .optimize_kl import OptimizeVI, OptimizeVIState, optimize_kl
from .prior import (
    InvGammaPrior, LaplacePrior, LogNormalPrior, NormalPrior, UniformPrior
)
from .refine.chart import CoordinateChart, HEALPixChart
from .refine.charted_field import RefinementField
from .refine.healpix_field import RefinementHPField
from .tree_math import *



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest

pmp = pytest.mark.parametrize


def test_grid_points():
    res = 64
    vol = 2
    sp = ift.RGSpace([res, res], [vol / res, vol / res])
    mg = np.mgrid[(slice(0, res),) * 2]
    mg = np.array(list(map(np.ravel, mg)))

    dist = [list(sp.distances)]
    dist = np.array(dist).reshape(-1, 1)

    sampling_points = dist * mg
    R = ift.LinearInterpolator(sp, sampling_points)

    ift.extra.check_linear_operator(R, atol=1e-7, rtol=1e-7)
    inp = ift.from_random(R.domain)
    out = R(inp).val
    np.testing.assert_allclose(out, inp.val.reshape(-1), rtol=1e-7)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from numpy.testing import assert_allclose

from ..common import list2fixture, setup_function, teardown_function


def _get_rtol(tp):
    if (tp == np.float64) or (tp == np.complex128):
        return 1e-10
    else:
        return 1e-5


space = list2fixture([ift.RGSpace(128)])
sigma = list2fixture([0., .5, 5.])
tp = list2fixture([np.float64, np.complex128])
pmp = pytest.mark.parametrize


def test_property(space, sigma):
    op = ift.HarmonicSmoothingOperator(space, sigma=sigma)
    if op.domain[0] != space:
        raise TypeError


def test_adjoint_times(space, sigma):
    op = ift.HarmonicSmoothingOperator(space, sigma=sigma)
    rand1 = ift.Field.from_random(domain=space, random_type='normal')
    rand2 = ift.Field.from_random(domain=space, random_type='normal')
    tt1 = rand1.s_vdot(op.times(rand2))
    tt2 = rand2.s_vdot(op.adjoint_times(rand1))
    assert_allclose(tt1, tt2)


def test_times(space, sigma):
    op = ift.HarmonicSmoothingOperator(space, sigma=sigma)
    fld = np.zeros(space.shape, dtype=np.float64)
    fld[0] = 1.
    rand1 = ift.Field.from_raw(space, fld)
    tt1 = op.times(rand1)
    assert_allclose(1, tt1.s_sum())


@pmp('sz', [128, 256])
@pmp('d', [1, 0.4])
def test_smooth_regular1(sz, d, sigma, tp):
    tol = _get_rtol(tp)
    sp = ift.RGSpace(sz, distances=d)
    smo = ift.HarmonicSmoothingOperator(sp, sigma=sigma)
    inp = ift.Field.from_random(domain=sp, random_type='normal', dtype=tp, std=1, mean=4)
    out = smo(inp)
    assert_allclose(inp.s_sum(), out.s_sum(), rtol=tol, atol=tol)


@pmp('sz1', [10, 15])
@pmp('sz2', [7, 10])
@pmp('d1', [1, 0.4])
@pmp('d2', [2, 0.3])
def test_smooth_regular2(sz1, sz2, d1, d2, sigma, tp):
    tol = _get_rtol(tp)
    sp = ift.RGSpace([sz1, sz2], distances=[d1, d2])
    smo = ift.HarmonicSmoothingOperator(sp, sigma=sigma)
    inp = ift.Field.from_random(domain=sp, random_type='normal', dtype=tp, std=1, mean=4)
    out = smo(inp)
    assert_allclose(inp.s_sum(), out.s_sum(), rtol=tol, atol=tol)



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from jax import numpy as jnp
import numpy as np
from numpy.testing import assert_allclose, assert_array_equal
import pytest

from nifty8.re.num import amend_unique, amend_unique_, unique

pmp = pytest.mark.parametrize


def _unique_via_amend(
    ar, *, axis, return_inverse=True, test_underscore=True, **kwargs
):
    assert return_inverse
    u = np.take(ar, (0, ), axis=axis)
    inv = [0]
    for el in np.moveaxis(ar, axis, 0)[1:]:
        u, i = amend_unique(u, el, **kwargs)
        inv += [i]

    if test_underscore:
        u_ = jnp.full_like(ar, jnp.nan)
        inv_ = []
        for el in np.moveaxis(ar, axis, 0):
            u_, i = amend_unique_(u_, el, **kwargs)
            inv_ += [i]
        # Remove NaN placeholder slices
        u_ = jnp.take(u_, np.arange(np.unique(inv_).size), axis=axis)
        assert_array_equal(u_, u)
        assert_array_equal(inv_, inv)

    return u, np.array(inv)


def test_unique_known_in_out():
    t0 = np.arange(4).reshape(2, 2)
    t1 = 1 + t0
    t1_1 = (1. + 1e-8) + t0
    t2 = 2. + t0
    test = np.stack([t0, t0, t1, t1, t1_1, t2, t1_1, t1, t0], axis=-1)

    u = unique(test, axis=-1)
    assert u.shape[-1] == 3
    au, _ = _unique_via_amend(test, axis=-1)
    assert au.shape[-1] == 3

    atol, rtol = 1e-12, 1e-12
    u, i = unique(test, atol=atol, rtol=rtol, axis=-1, return_inverse=True)
    assert u.shape[-1] == 4
    assert_allclose(
        test, np.take(u, i, axis=-1), atol=1.01 * atol, rtol=1.01 * rtol
    )
    au, ai = _unique_via_amend(
        test, atol=atol, rtol=rtol, axis=-1, return_inverse=True
    )
    assert_array_equal(au, u)
    assert_array_equal(ai, i)


@pmp("seed", (42, 43))
@pmp("shape", ((5, ), (8, 8, 13), (2, 13)))
def test_unique_fuzz(seed, shape):
    sseq = np.random.SeedSequence(seed)
    rng = np.random.default_rng(sseq)
    atol, rtol = 1e-5, 1e-4

    d = rng.normal(size=shape)
    for axis in range(d.ndim):
        da = np.repeat(d, rng.integers(1, 5, size=d.shape[axis]), axis=axis)
        da = rng.permutation(da, axis=axis)
        u, i = unique(da, atol=atol, rtol=rtol, axis=-1, return_inverse=True)
        assert_allclose(
            da,
            np.take(u, i, axis=-1),
            atol=1.01 * atol,
            rtol=1.01 * rtol,
            equal_nan=False
        )
        au, ai = _unique_via_amend(
            da, atol=atol, rtol=rtol, axis=-1, return_inverse=True
        )
        assert_array_equal(au, u)
        assert_array_equal(ai, i)


if __name__ == "__main__":
    test_unique_known_in_out()
    test_unique_fuzz(42, (8, 8, 13))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import collections
import pickle
from functools import reduce
from itertools import product

import numpy as np

__all__ = ["get_slice_list", "safe_cast", "parse_spaces", "infer_space",
           "memo", "NiftyMeta", "my_sum", "my_lincomb_simple",
           "my_lincomb", "indent", "my_product", "frozendict",
           "special_add_at", "iscomplextype", "issingleprec",
           "value_reshaper", "lognormal_moments", "check_object_identity",
           "check_MPI_equality", "check_MPI_synced_random_state"]


def my_sum(iterable):
    return reduce(lambda x, y: x+y, iterable)


def my_lincomb_simple(terms, factors):
    terms2 = map(lambda v: v[0]*v[1], zip(terms, factors))
    return my_sum(terms2)


def my_lincomb(terms, factors):
    terms2 = map(lambda v: v[0] if v[1] == 1. else v[0]*v[1],
                 zip(terms, factors))
    return my_sum(terms2)


def my_product(iterable):
    return reduce(lambda x, y: x*y, iterable)


def get_slice_list(shape, axes):
    """
    Helper function which generates slice list(s) to traverse over all
    combinations of axes, other than the selected axes.

    Parameters
    ----------
    shape: tuple
        Shape of the data array to traverse over.
    axes: tuple
        Axes which should not be iterated over.

    Yields
    ------
    list
        The next list of indices and/or slice objects for each dimension.

    Raises
    ------
    ValueError
        If shape is empty.
        If axes(axis) does not match shape.
    """
    if shape is None:
        raise ValueError("shape cannot be None.")

    if axes:
        if not all(axis < len(shape) for axis in axes):
            raise ValueError("axes(axis) does not match shape.")
        axes_select = [0 if x in axes else 1 for x in range(len(shape))]
        axes_iterables = \
            [list(range(y)) for x, y in enumerate(shape) if x not in axes]
        for index in product(*axes_iterables):
            it_iter = iter(index)
            slice_list = tuple(
                next(it_iter)
                if axis else slice(None, None) for axis in axes_select
            )
            yield slice_list
    else:
        yield [slice(None, None)]


def safe_cast(tfunc, val):
    tmp = tfunc(val)
    if val != tmp:
        raise ValueError("value changed during cast")
    return tmp


def parse_spaces(spaces, nspc):
    nspc = safe_cast(int, nspc)
    if spaces is None:
        return tuple(range(nspc))
    elif np.isscalar(spaces):
        spaces = (safe_cast(int, spaces),)
    else:
        spaces = tuple(safe_cast(int, item) for item in spaces)
    if len(spaces) == 0:
        return spaces
    tmp = tuple(set(spaces))
    if tmp[0] < 0 or tmp[-1] >= nspc:
        raise ValueError("space index out of range")
    if len(tmp) != len(spaces):
        raise ValueError("multiply defined space indices")
    return spaces


def infer_space(domain, space):
    if space is None:
        if len(domain) != 1:
            raise ValueError("'space' index must be given for objects based on"
                             " DomainTuples containing more than one domain")
        space = 0
    space = int(space)
    if space < 0 or space >= len(domain):
        raise ValueError("space index out of range")
    return space


def memo(f):
    name = f.__name__

    def wrapped_f(self):
        if not hasattr(self, "_cache"):
            self._cache = {}
        try:
            return self._cache[name]
        except KeyError:
            self._cache[name] = f(self)
            return self._cache[name]
    return wrapped_f


class _DocStringInheritor(type):
    """
    A variation on
    https://groups.google.com/group/comp.lang.python/msg/26f7b4fcb4d66c95
    by Paul McGuire
    """
    def __new__(meta, name, bases, clsdict):
        if not('__doc__' in clsdict and clsdict['__doc__']):
            for mro_cls in (mro_cls for base in bases
                            for mro_cls in base.mro()):
                doc = mro_cls.__doc__
                if doc:
                    clsdict['__doc__'] = doc
                    break
        for attr, attribute in clsdict.items():
            if not attribute.__doc__:
                for mro_cls in (mro_cls for base in bases
                                for mro_cls in base.mro()
                                if hasattr(mro_cls, attr)):
                    doc = getattr(getattr(mro_cls, attr), '__doc__')
                    if doc:
                        if isinstance(attribute, property):
                            clsdict[attr] = property(attribute.fget,
                                                     attribute.fset,
                                                     attribute.fdel,
                                                     doc)
                        else:
                            attribute.__doc__ = doc
                        break
        return super(_DocStringInheritor, meta).__new__(meta, name,
                                                        bases, clsdict)


class NiftyMeta(_DocStringInheritor):
    pass


class frozendict(collections.abc.Mapping):
    """
    An immutable wrapper around dictionaries that implements the complete
    :py:class:`collections.Mapping` interface. It can be used as a drop-in
    replacement for dictionaries where immutability is desired.
    """

    dict_cls = dict

    def __init__(self, *args, **kwargs):
        self._dict = self.dict_cls(*args, **kwargs)
        self._hash = None

    def __getitem__(self, key):
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict

    def copy(self, **add_or_replace):
        return self.__class__(self, **add_or_replace)

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self._dict)

    def __hash__(self):
        if self._hash is None:
            h = 0
            for key, value in self._dict.items():
                h ^= hash((key, value))
            self._hash = h
        return self._hash


def special_add_at(a, axis, index, b):
    if a.dtype != b.dtype:
        raise TypeError("data type mismatch")
    sz1 = int(np.prod(a.shape[:axis]))
    sz3 = int(np.prod(a.shape[axis+1:]))
    a2 = a.reshape([sz1, -1, sz3])
    b2 = b.reshape([sz1, -1, sz3])
    if iscomplextype(a.dtype):
        dt2 = a.real.dtype
        a2 = a2.view(dt2)
        b2 = b2.view(dt2)
        sz3 *= 2
    for i1 in range(sz1):
        for i3 in range(sz3):
            a2[i1, :, i3] += np.bincount(index, b2[i1, :, i3],
                                         minlength=a2.shape[1])

    if iscomplextype(a.dtype):
        a2 = a2.view(a.dtype)
    return a2.reshape(a.shape)


def iscomplextype(dtype):
    if isinstance(dtype, dict):
        return _getunique(iscomplextype, dtype.values())
    return np.issubdtype(dtype, np.complexfloating)


def issingleprec(dtype):
    if isinstance(dtype, dict):
        return _getunique(issingleprec, dtype.values())
    return dtype.type in (np.float32, np.complex64)


def _getunique(f, iterable):
    lst = list(f(vv) for vv in iterable)
    if len(set(lst)) == 1:
        return lst[0]
    raise RuntimeError("Value is not unique", lst)


def indent(inp):
    return "\n".join((("  "+s).rstrip() for s in inp.splitlines()))


def shareRange(nwork, nshares, myshare):
    """Divides a number of work items as fairly as possible into a given number
    of shares.

    Parameters
    ----------
    nwork: int
        number of work items
    nshares: int
        number of shares among which the work should be distributed
    myshare: int
        the share for which the range of work items is requested


    Returns
    -------
    lo, hi: int
        index range of work items for this share
    """

    nbase = nwork//nshares
    additional = nwork % nshares
    lo = myshare*nbase + min(myshare, additional)
    hi = lo + nbase + int(myshare < additional)
    return lo, hi


def get_MPI_params_from_comm(comm):
    if comm is None:
        return 1, 0, True
    size = comm.Get_size()
    rank = comm.Get_rank()
    return size, rank, rank == 0


def get_MPI_params():
    """Returns basic information about the MPI setup of the running script.

    Returns
    -------
    comm: MPI communicator or None
        if MPI is detected _and_ more than one task is active, returns
        the world communicator, else returns None
    size: int
        the number of tasks running in total
    rank: int
        the rank of this task
    master: bool
        True if rank == 0, else False
    """

    try:
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        size = comm.Get_size()
        if size == 1:
            return None, 1, 0, True
        rank = comm.Get_rank()
        return comm, size, rank, rank == 0
    except ImportError:
        return None, 1, 0, True


def allreduce_sum(obj, comm):
    """ This is a deterministic implementation of MPI allreduce

    Numeric addition is not associative due to rounding errors.
    Therefore we provide our own implementation that is consistent
    no matter if MPI is used and how many tasks there are.

    At the beginning, a list `who` is constructed, that states which obj can
    be found on which MPI task.
    Then elements are added pairwise, with increasing pair distance.
    In the first round, the distance between pair members is 1:
      v[0] := v[0] + v[1]
      v[2] := v[2] + v[3]
      v[4] := v[4] + v[5]
    Entries whose summation partner lies beyond the end of the array
    stay unchanged.
    When both summation partners are not located on the same MPI task,
    the second summand is sent to the task holding the first summand and
    the operation is carried out there.
    For the next round, the distance is doubled:
      v[0] := v[0] + v[2]
      v[4] := v[4] + v[6]
      v[8] := v[8] + v[10]
    This is repeated until the distance exceeds the length of the array.
    At this point v[0] contains the sum of all entries, which is then
    broadcast to all tasks.
    """
    vals = list(obj)
    if comm is None:
        nobj = len(vals)
        who = np.zeros(nobj, dtype=np.int32)
        rank = 0
    else:
        rank = comm.Get_rank()
        nobj_list = comm.allgather(len(vals))
        all_hi = list(np.cumsum(nobj_list))
        all_lo = [0] + all_hi[:-1]
        nobj = all_hi[-1]
        rank_lo_hi = [(l, h) for l, h in zip(all_lo, all_hi)]
        lo, hi = rank_lo_hi[rank]
        vals = [None]*lo + vals + [None]*(nobj-hi)
        who = [t for t, (l, h) in enumerate(rank_lo_hi) for cnt in range(h-l)]

    step = 1
    while step < nobj:
        for j in range(0, nobj, 2*step):
            if j+step < nobj:  # summation partner found
                if rank == who[j]:
                    if who[j] == who[j+step]:  # no communication required
                        vals[j] = vals[j] + vals[j+step]
                        vals[j+step] = None
                    else:
                        vals[j] = vals[j] + comm.recv(source=who[j+step])
                elif rank == who[j+step]:
                    comm.send(vals[j+step], dest=who[j])
                    vals[j+step] = None
        step *= 2
    if comm is None:
        return vals[0]
    return comm.bcast(vals[0], root=who[0])


def value_reshaper(x, N):
    """Produce arrays of shape `(N,)`.
    If `x` is a scalar or array of length one, fill the target array with it.
    If `x` is an array, check if it has the right shape."""
    x = np.asarray(x, dtype=float)
    if x.shape in [(), (1, )]:
        return np.full(N, x) if N != 0 else x.reshape(())
    elif x.shape == (N, ):
        return x
    raise TypeError("x and N are incompatible")


def lognormal_moments(mean, sigma, N=0):
    """Calculates the parameters for a normal distribution `n(x)`
    such that `exp(n)(x)` has the mean and standard deviation given.

    Used in :func:`~nifty8.normal_operators.LognormalTransform`."""
    mean, sigma = (value_reshaper(param, N) for param in (mean, sigma))
    if not np.all(mean > 0):
        raise ValueError("mean must be greater 0; got {!r}".format(mean))
    if not np.all(sigma > 0):
        raise ValueError("sig must be greater 0; got {!r}".format(sigma))

    logsigma = np.sqrt(np.log1p((sigma / mean)**2))
    logmean = np.log(mean) - logsigma**2 / 2
    return logmean, logsigma


def myassert(val):
    """Safe alternative to python's assert statement which is active even if
    `__debug__` is False."""
    if not val:
        raise AssertionError


def check_object_identity(obj0, obj1):
    """Check if two objects are the same and throw ValueError if not."""
    if obj0 is not obj1:
        raise ValueError(f"Mismatch:\n{obj0}\n{obj1}")


def check_MPI_equality(obj, comm, hash=False):
    """Check that object is the same on all MPI tasks associated to a given
    communicator.

    Raises a RuntimeError if it differs.

    Parameters
    ----------
    obj :
        Any Python object that implements __eq__.
    comm : MPI communicator or None
        If comm is None, no check will be performed
    """
    if comm is None:
        return
    if not _MPI_unique(obj, comm, hash=hash):
        raise RuntimeError("MPI tasks are not in sync")


def _MPI_unique(obj, comm, hash=False):
    from hashlib import blake2b

    obj = pickle.dumps(obj)
    obj = blake2b(obj).hexdigest() if hash else obj
    return len(set(comm.allgather(obj))) == 1


def check_MPI_synced_random_state(comm):
    """Check that random state is the same on all MPI tasks associated to a
    given communicator.

    Raises a RuntimeError if it differs.

    Parameters
    ----------
    comm : MPI communicator or None
        If comm is None, no check will be performed
    """
    from .random import getState
    if comm is None:
        return
    check_MPI_equality(getState(), comm)


def check_dtype_or_none(obj, domain=None):
    """Check that dtype is compatible with a given domain.

    If domain is None or a DomainTuple, the obj is checked to be either a
    floating dtype or None. If domain is a MultiDomain, obj can still be a
    single quantity and is treated like in the previous case. Or it can be a
    dictionary; then all its entries need to satisfy the previous condition to
    pass the test.

    Raises a TypeError if any incompatibility is detected.

    Parameters
    ----------
    obj : object
        The object to be checked.
    domain : DomainTuple or MultiDomain
        If it is a MultiDomain,
    """
    from .multi_domain import MultiDomain
    from .sugar import makeDomain
    if domain is not None:
        domain = makeDomain(domain)
        if isinstance(domain, MultiDomain) and isinstance(obj, dict):
            for kk in domain.keys():
                check_dtype_or_none(obj[kk])
            return
    check = obj in [np.float32, np.float64, float,
                    np.complex64, np.complex128, complex,
                    None]
    if not check:
        s = "Need to pass floating dtype (e.g. np.float64, complex) "
        s += f"or `None` to this function.\nHave recieved:\n{obj}"
        raise TypeError(s)


class Nop:
    def nop(*args, **kw):
        return Nop()
    def __getattr__(self, _):
        return self.nop



#!/usr/bin/env python3

# Copyright(C) 2013-2021 Max-Planck-Society
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

# %% [markdown]
# # 3D Tomography Example with known start- and end-points

# %%
import jax
import matplotlib.pyplot as plt
import numpy as np
from jax import numpy as jnp
from jax import random

import nifty8.re as jft

# %%
key = random.PRNGKey(42)

dims = (64, 64, 64)
distances = tuple(1.0 / d for d in dims)
cfm = jft.CorrelatedFieldMaker("cf")
cfm.set_amplitude_total_offset(offset_mean=2, offset_std=(1e-1, 3e-2))
cfm.add_fluctuations(  # Axis over which the kernle is defined
    dims,
    distances=distances,
    fluctuations=(1.0, 5e-1),
    loglogavgslope=(-5.0, 2e-1),
    flexibility=(1e0, 2e-1),
    asperity=(5e-1, 5e-2),
    prefix="ax1",
    non_parametric_kind="power",
)
correlated_field = cfm.finalize()  # forward model for a GP prior


# %%
# Draw synthetic end points
n_synth_points = 100
key, sk = random.split(key)
end = random.uniform(
    sk,
    (n_synth_points, len(dims)),
    minval=0.05,
    maxval=0.95,
)

# %%
start = (0.5, 0.5, 0.5)
n_sampling_points = 256
los = jft.SamplingCartesianGridLOS(
    start,
    end,
    distances=distances,
    shape=correlated_field.target.shape,
    dtype=correlated_field.target.dtype,
    n_sampling_points=n_sampling_points,
)


class Forward(jft.Model):
    def __init__(self, log_density, los):
        self.log_density = log_density
        self.los = los
        # Track a method with which a random input for the model. This is not
        # strictly required but is usually handy when building deep models.
        super().__init__(init=log_density.init)

    def density(self, x):
        return jnp.exp(self.log_density(x))

    def __call__(self, x):
        return self.los(self.density(x))


forward = Forward(correlated_field, los)

# %%
key, k_f, k_n = random.split(key, 3)
synth_pos = forward.init(k_f)
synth_truth = forward(synth_pos)
noise_scl = 0.15
synth_noise = random.normal(k_n, synth_truth.shape, synth_truth.dtype)
synth_noise = synth_noise * noise_scl * synth_truth
synth_data = synth_truth + synth_noise

# %%
lh = jft.Gaussian(synth_data, noise_cov_inv=1.0 / synth_noise**2).amend(forward)

# %%
n_vi_iterations = 6
delta = 1e-4
n_samples = 4

key, k_i, k_o = random.split(key, 3)
samples, state = jft.optimize_kl(
    lh,
    jft.Vector(lh.init(k_i)),
    n_total_iterations=n_vi_iterations,
    n_samples=lambda i: n_samples // 2 if i < 2 else n_samples,
    key=k_o,
    draw_linear_kwargs=dict(
        cg_name="SL",
        cg_kwargs=dict(absdelta=delta * jft.size(lh.domain) / 10.0, maxiter=100),
    ),
    nonlinearly_update_kwargs=dict(
        minimize_kwargs=dict(
            name="SN",
            xtol=delta,
            cg_kwargs=dict(name=None),
            maxiter=5,
        )
    ),
    kl_kwargs=dict(
        minimize_kwargs=dict(
            name="M", xtol=delta, cg_kwargs=dict(name="MCG"), maxiter=35
        )
    ),
    sample_mode="linear_resample",
    odir="results_tomography",
    resume=False,
)

# %%
synth_density = forward.density(synth_pos)
post_density = jax.vmap(forward.density)(samples.samples)

# %%
extent = (0, 1, 0, 1)
fig, axs = plt.subplots(3, 3, dpi=500)
for i, (ax_t, ax_p, ax_ps) in enumerate(axs):
    ax_t.imshow(synth_density.sum(axis=i), extent=extent)
    not_i = tuple(set.difference({0, 1, 2}, {i}))
    ax_t.plot(
        *end[:, not_i].T,
        "+",
        markersize=0.5,
        color="red",
    )
    ax_t.plot(
        *np.atleast_2d(start)[:, not_i].T,
        ".",
        markersize=6,
        color="red",
    )
    ax_p.imshow(post_density.mean(axis=0).sum(axis=i), extent=extent)
    ax_ps.imshow(post_density.std(axis=0).sum(axis=i), extent=extent)
for title, ax in zip(("Truth", "Posterior mean", "Posterior std."), axs[0]):
    ax.set_title(title)
fig.savefig("results_tomography_projected_full_reconstruction.png", dpi=400)
plt.show()



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.13.7
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# + [markdown] slideshow={"slide_type": "slide"}
# # Code example: Wiener filter

# + [markdown] slideshow={"slide_type": "subslide"}
# ## Introduction
# IFT starting point:
#
# $$d = Rs+n$$
#
# Typically, $s$ is a continuous field, $d$ a discrete data vector. Particularly, $R$ is not invertible.
#
# IFT aims at **inverting** the above uninvertible problem in the **best possible way** using Bayesian statistics.
#
# NIFTy (Numerical Information Field Theory) is a Python framework in which IFT problems can be tackled easily.
#
# Main Interfaces:
#
# - **Spaces**: Cartesian, 2-Spheres (Healpix, Gauss-Legendre) and their respective harmonic spaces.
# - **Fields**: Defined on spaces.
# - **Operators**: Acting on fields.

# + [markdown] slideshow={"slide_type": "subslide"}
# ## Wiener filter on one-dimensional fields
#
# ### Assumptions
#
# - $d=Rs+n$, $R$ linear operator.
# - $\mathcal P (s) = \mathcal G (s,S)$, $\mathcal P (n) = \mathcal G (n,N)$ where $S, N$ are positive definite matrices.
#
# ### Posterior
# The Posterior is given by:
#
# $$\mathcal P (s|d) \propto P(s,d) = \mathcal G(d-Rs,N) \,\mathcal G(s,S) \propto \mathcal G (s-m,D) $$
#
# where
# $$m = Dj$$
# with
# $$D = (S^{-1} +R^\dagger N^{-1} R)^{-1} , \quad j = R^\dagger N^{-1} d.$$
#
# Let us implement this in NIFTy!

# + [markdown] slideshow={"slide_type": "subslide"}
# ### In NIFTy
#
# - We assume statistical homogeneity and isotropy. Therefore the signal covariance $S$ is diagonal in harmonic space, and is described by a one-dimensional power spectrum, assumed here as $$P(k) = P_0\,\left(1+\left(\frac{k}{k_0}\right)^2\right)^{-\gamma /2},$$
# with $P_0 = 0.2, k_0 = 5, \gamma = 4$.
# - $N = 0.2 \cdot \mathbb{1}$.
# - Number of data points $N_{pix} = 512$.
# - reconstruction in harmonic space.
# - Response operator:
# $$R = FFT_{\text{harmonic} \rightarrow \text{position}}$$
#

# + slideshow={"slide_type": "-"}
N_pixels = 512     # Number of pixels

def pow_spec(k):
    P0, k0, gamma = [.2, 5, 4]
    return P0 / ((1. + (k/k0)**2)**(gamma / 2))


# + [markdown] slideshow={"slide_type": "slide"}
# ### Implementation

# + [markdown] slideshow={"slide_type": "-"}
# #### Import Modules

# + slideshow={"slide_type": "-"}
# %matplotlib inline
import numpy as np
import nifty8 as ift
import matplotlib.pyplot as plt

plt.rcParams['figure.dpi'] = 100


# + [markdown] slideshow={"slide_type": "subslide"}
# #### Implement Propagator

# + slideshow={"slide_type": "-"}
def Curvature(R, N, Sh):
    IC = ift.GradientNormController(iteration_limit=50000,
                                    tol_abs_gradnorm=0.1)
    # WienerFilterCurvature is (R.adjoint*N.inverse*R + Sh.inverse) plus some handy
    # helper methods.
    return ift.WienerFilterCurvature(R,N,Sh,iteration_controller=IC,iteration_controller_sampling=IC)


# + [markdown] slideshow={"slide_type": "skip"}
# #### Conjugate Gradient Preconditioning
#
# - $D$ is defined via:
# $$D^{-1} = \mathcal S_h^{-1} + R^\dagger N^{-1} R.$$
# In the end, we want to apply $D$ to $j$, i.e. we need the inverse action of $D^{-1}$. This is done numerically (algorithm: *Conjugate Gradient*). 
#
# <!--
# - One can define the *condition number* of a non-singular and normal matrix $A$:
# $$\kappa (A) := \frac{|\lambda_{\text{max}}|}{|\lambda_{\text{min}}|},$$
# where $\lambda_{\text{max}}$ and $\lambda_{\text{min}}$ are the largest and smallest eigenvalue of $A$, respectively.
#
# - The larger $\kappa$ the slower Conjugate Gradient.
#
# - By default, conjugate gradient solves: $D^{-1} m = j$ for $m$, where $D^{-1}$ can be badly conditioned. If one knows a non-singular matrix $T$ for which $TD^{-1}$ is better conditioned, one can solve the equivalent problem:
# $$\tilde A m = \tilde j,$$
# where $\tilde A = T D^{-1}$ and $\tilde j = Tj$.
#
# - In our case $S^{-1}$ is responsible for the bad conditioning of $D$ depending on the chosen power spectrum. Thus, we choose
#
# $$T = \mathcal F^\dagger S_h^{-1} \mathcal F.$$
# -->

# + [markdown] slideshow={"slide_type": "subslide"}
# #### Generate Mock data
#
# - Generate a field $s$ and $n$ with given covariances.
# - Calculate $d$.

# +
s_space = ift.RGSpace(N_pixels)
h_space = s_space.get_default_codomain()
HT = ift.HarmonicTransformOperator(h_space, target=s_space)

# Operators
Sh = ift.create_power_operator(h_space, power_spectrum=pow_spec, sampling_dtype=float)
R = HT # @ ift.create_harmonic_smoothing_operator((h_space,), 0, 0.02)

# Fields and data
sh = Sh.draw_sample()
noiseless_data=R(sh)
noise_amplitude = np.sqrt(0.2)
N = ift.ScalingOperator(s_space, noise_amplitude**2, float)

n = ift.Field.from_random(domain=s_space, random_type='normal',
                          std=noise_amplitude, mean=0)
d = noiseless_data + n
j = R.adjoint_times(N.inverse_times(d))
curv = Curvature(R=R, N=N, Sh=Sh)
D = curv.inverse

# + [markdown] slideshow={"slide_type": "subslide"}
# #### Run Wiener Filter

# + slideshow={"slide_type": "-"}
m = D(j)

# + [markdown] slideshow={"slide_type": "subslide"}
# #### Results

# + slideshow={"slide_type": "-"}
# Get signal data and reconstruction data
s_data = HT(sh).val
m_data = HT(m).val
d_data = d.val

plt.plot(s_data, 'r', label="Signal", linewidth=2)
plt.plot(d_data, 'k.', label="Data")
plt.plot(m_data, 'k', label="Reconstruction",linewidth=2)
plt.title("Reconstruction")
plt.legend()
plt.show()

# + slideshow={"slide_type": "subslide"}
plt.plot(s_data - s_data, 'r', label="Signal", linewidth=2)
plt.plot(d_data - s_data, 'k.', label="Data")
plt.plot(m_data - s_data, 'k', label="Reconstruction",linewidth=2)
plt.axhspan(-noise_amplitude,noise_amplitude, facecolor='0.9', alpha=.5)
plt.title("Residuals")
plt.legend()
plt.show()

# + [markdown] slideshow={"slide_type": "subslide"}
# #### Power Spectrum

# + slideshow={"slide_type": "-"}
s_power_data = ift.power_analyze(sh).val
m_power_data = ift.power_analyze(m).val
plt.loglog()
plt.xlim(1, int(N_pixels/2))
ymin = min(m_power_data)
plt.ylim(ymin, 1)
xs = np.arange(1,int(N_pixels/2),.1)
plt.plot(xs, pow_spec(xs), label="True Power Spectrum", color='k',alpha=0.5)
plt.plot(s_power_data, 'r', label="Signal")
plt.plot(m_power_data, 'k', label="Reconstruction")
plt.axhline(noise_amplitude**2 / N_pixels, color="k", linestyle='--', label="Noise level", alpha=.5)
plt.axhspan(noise_amplitude**2 / N_pixels, ymin, facecolor='0.9', alpha=.5)
plt.title("Power Spectrum")
plt.legend()
plt.show()

# + [markdown] slideshow={"slide_type": "slide"}
# ## Wiener Filter on Incomplete Data

# + slideshow={"slide_type": "skip"}
# Operators
Sh = ift.create_power_operator(h_space, power_spectrum=pow_spec, sampling_dtype=float)
N = ift.ScalingOperator(s_space, noise_amplitude**2, sampling_dtype=float)
# R is defined below

# Fields
sh = Sh.draw_sample()
s = HT(sh)
n = ift.Field.from_random(domain=s_space, random_type='normal',
                      std=noise_amplitude, mean=0)

# + [markdown] slideshow={"slide_type": "skip"}
# ### Partially Lose Data

# + slideshow={"slide_type": "-"}
l = int(N_pixels * 0.2)
h = int(N_pixels * 0.2 * 2)

mask = np.full(s_space.shape, 1.)
mask[l:h] = 0
mask = ift.Field.from_raw(s_space, mask)

R = ift.DiagonalOperator(mask) @ HT
n = n.val_rw()
n[l:h] = 0
n = ift.Field.from_raw(s_space, n)

d = R(sh) + n

# + slideshow={"slide_type": "skip"}
curv = Curvature(R=R, N=N, Sh=Sh)
D = curv.inverse
j = R.adjoint_times(N.inverse_times(d))
m = D(j)

# + [markdown] slideshow={"slide_type": "subslide"}
# ### Compute Uncertainty
#
# -

m_mean, m_var = ift.probe_with_posterior_samples(curv, HT, 200, np.float64)

# + [markdown] slideshow={"slide_type": "skip"}
# ### Get data

# + slideshow={"slide_type": "skip"}
# Get signal data and reconstruction data
s_data = s.val
m_data = HT(m).val
m_var_data = m_var.val
uncertainty = np.sqrt(m_var_data)
d_data = d.val_rw()

# Set lost data to NaN for proper plotting
d_data[d_data == 0] = np.nan

# + slideshow={"slide_type": "skip"}
plt.axvspan(l, h, facecolor='0.8',alpha=0.5)
plt.fill_between(range(N_pixels), m_data - uncertainty, m_data + uncertainty, facecolor='0.5', alpha=0.5)
plt.plot(s_data, 'r', label="Signal", alpha=1, linewidth=2)
plt.plot(d_data, 'k.', label="Data")
plt.plot(m_data, 'k', label="Reconstruction", linewidth=2)
plt.title("Reconstruction of incomplete data")
plt.legend()

# + [markdown] slideshow={"slide_type": "slide"}
# ## Wiener filter on two-dimensional field

# +
N_pixels = 256      # Number of pixels
sigma2 = 2.         # Noise variance

def pow_spec(k):
    P0, k0, gamma = [.2, 2, 4]
    return P0 * (1. + (k/k0)**2)**(-gamma/2)

s_space = ift.RGSpace([N_pixels, N_pixels])

# + slideshow={"slide_type": "skip"}
h_space = s_space.get_default_codomain()
HT = ift.HarmonicTransformOperator(h_space,s_space)

# Operators
Sh = ift.create_power_operator(h_space, power_spectrum=pow_spec, sampling_dtype=float)
N = ift.ScalingOperator(s_space, sigma2, sampling_dtype=float)

# Fields and data
sh = Sh.draw_sample()
n = ift.Field.from_random(domain=s_space, random_type='normal',
                      std=np.sqrt(sigma2), mean=0)

# Lose some data

l = int(N_pixels * 0.33)
h = int(N_pixels * 0.33 * 2)

mask = np.full(s_space.shape, 1.)
mask[l:h,l:h] = 0.
mask = ift.Field.from_raw(s_space, mask)

R = ift.DiagonalOperator(mask)(HT)
n = n.val_rw()
n[l:h, l:h] = 0
n = ift.Field.from_raw(s_space, n)
curv = Curvature(R=R, N=N, Sh=Sh)
D = curv.inverse

d = R(sh) + n
j = R.adjoint_times(N.inverse_times(d))

# Run Wiener filter
m = D(j)

# Uncertainty
m_mean, m_var = ift.probe_with_posterior_samples(curv, HT, 20, np.float64)

# Get data
s_data = HT(sh).val
m_data = HT(m).val
m_var_data = m_var.val
d_data = d.val
uncertainty = np.sqrt(np.abs(m_var_data))

# + slideshow={"slide_type": "skip"}
cmap = ['magma', 'inferno', 'plasma', 'viridis'][1]

mi = np.min(s_data)
ma = np.max(s_data)

fig, axes = plt.subplots(1, 2)

data = [s_data, d_data]
caption = ["Signal", "Data"]

for ax in axes.flat:
    im = ax.imshow(data.pop(0), interpolation='nearest', cmap=cmap, vmin=mi,
                   vmax=ma)
    ax.set_title(caption.pop(0))

fig.subplots_adjust(right=0.8)
cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
fig.colorbar(im, cax=cbar_ax)

# + slideshow={"slide_type": "skip"}
mi = np.min(s_data)
ma = np.max(s_data)

fig, axes = plt.subplots(3, 2, figsize=(10, 15))
sample = HT(curv.draw_sample(from_inverse=True)+m).val
post_mean = (m_mean + HT(m)).val

data = [s_data, m_data, post_mean, sample, s_data - m_data, uncertainty]
caption = ["Signal", "Reconstruction", "Posterior mean", "Sample", "Residuals", "Uncertainty Map"]

for ax in axes.flat:
    im = ax.imshow(data.pop(0), interpolation='nearest', cmap=cmap, vmin=mi, vmax=ma)
    ax.set_title(caption.pop(0))

fig.subplots_adjust(right=0.8)
cbar_ax = fig.add_axes([.85, 0.15, 0.05, 0.7])
fig.colorbar(im, cax=cbar_ax)

# + [markdown] slideshow={"slide_type": "subslide"}
# ### Is the uncertainty map reliable?

# + slideshow={"slide_type": "-"}
precise = (np.abs(s_data-m_data) < uncertainty)
print("Error within uncertainty map bounds: " + str(np.sum(precise) * 100 / N_pixels**2) + "%")

plt.imshow(precise.astype(float), cmap="brg")
plt.colorbar()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import pytest

from ..common import setup_function, teardown_function

pmp = pytest.mark.parametrize


@pmp('f', [
    'log', 'exp', 'sqrt', 'sin', 'cos', 'tan', 'sinc', 'sinh', 'cosh', 'tanh',
    'absolute', 'reciprocal', 'sigmoid', 'log10', 'log1p', 'expm1', 'softplus',
    ('power', 2.), ('exponentiate', 1.1)
])
def test_ptw_pre(f):
    if not isinstance(f, tuple):
        f = (f,)
    op = ift.FFTOperator(ift.RGSpace(10))
    op0 = op @ ift.ScalingOperator(op.domain, 1.).ptw(*f)
    op1 = op.ptw_pre(*f)
    pos = ift.from_random(op0.domain)
    if f[0] in ['log', 'sqrt', 'log10', 'log1p']:
        pos = pos.exp()
    ift.extra.assert_equal(op0(pos), op1(pos))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import os
from datetime import datetime as dt
from itertools import product
from warnings import warn

import numpy as np

from .domain_tuple import DomainTuple
from .domains.gl_space import GLSpace
from .domains.hp_space import HPSpace
from .domains.power_space import PowerSpace
from .domains.rg_space import RGSpace
from .field import Field
from .minimization.iteration_controllers import EnergyHistory
from .multi_field import MultiField
from .utilities import check_object_identity, myassert

# relevant properties:
# - x/y size
# - x/y/z log
# - x/y/z min/max
# - colorbar/colormap
# - axis on/off
# - title
# - axis labels
# - labels


def _mollweide_helper(xsize):
    xsize = int(xsize)
    ysize = xsize//2
    res = np.full(shape=(ysize, xsize), fill_value=np.nan, dtype=np.float64)
    xc, yc = (xsize-1)*0.5, (ysize-1)*0.5
    u, v = np.meshgrid(np.arange(xsize), np.arange(ysize))
    u, v = 2*(u-xc)/(xc/1.02), (v-yc)/(yc/1.02)

    mask = np.where((u*u*0.25 + v*v) <= 1.)
    t1 = v[mask]
    theta = 0.5*np.pi-(
        np.arcsin(2/np.pi*(np.arcsin(t1) + t1*np.sqrt((1.-t1)*(1+t1)))))
    phi = -0.5*np.pi*u[mask]/np.maximum(np.sqrt((1-t1)*(1+t1)), 1e-6)
    phi = np.where(phi < 0, phi+2*np.pi, phi)
    return res, mask, theta, phi


def _rgb_data(spectral_cube):
    _xyz = np.array(
          [[0.000160, 0.000662, 0.002362, 0.007242, 0.019110,
            0.043400, 0.084736, 0.140638, 0.204492, 0.264737,
            0.314679, 0.357719, 0.383734, 0.386726, 0.370702,
            0.342957, 0.302273, 0.254085, 0.195618, 0.132349,
            0.080507, 0.041072, 0.016172, 0.005132, 0.003816,
            0.015444, 0.037465, 0.071358, 0.117749, 0.172953,
            0.236491, 0.304213, 0.376772, 0.451584, 0.529826,
            0.616053, 0.705224, 0.793832, 0.878655, 0.951162,
            1.014160, 1.074300, 1.118520, 1.134300, 1.123990,
            1.089100, 1.030480, 0.950740, 0.856297, 0.754930,
            0.647467, 0.535110, 0.431567, 0.343690, 0.268329,
            0.204300, 0.152568, 0.112210, 0.081261, 0.057930,
            0.040851, 0.028623, 0.019941, 0.013842, 0.009577,
            0.006605, 0.004553, 0.003145, 0.002175, 0.001506,
            0.001045, 0.000727, 0.000508, 0.000356, 0.000251,
            0.000178, 0.000126, 0.000090, 0.000065, 0.000046,
            0.000033],
           [0.000017, 0.000072, 0.000253, 0.000769, 0.002004,
            0.004509, 0.008756, 0.014456, 0.021391, 0.029497,
            0.038676, 0.049602, 0.062077, 0.074704, 0.089456,
            0.106256, 0.128201, 0.152761, 0.185190, 0.219940,
            0.253589, 0.297665, 0.339133, 0.395379, 0.460777,
            0.531360, 0.606741, 0.685660, 0.761757, 0.823330,
            0.875211, 0.923810, 0.961988, 0.982200, 0.991761,
            0.999110, 0.997340, 0.982380, 0.955552, 0.915175,
            0.868934, 0.825623, 0.777405, 0.720353, 0.658341,
            0.593878, 0.527963, 0.461834, 0.398057, 0.339554,
            0.283493, 0.228254, 0.179828, 0.140211, 0.107633,
            0.081187, 0.060281, 0.044096, 0.031800, 0.022602,
            0.015905, 0.011130, 0.007749, 0.005375, 0.003718,
            0.002565, 0.001768, 0.001222, 0.000846, 0.000586,
            0.000407, 0.000284, 0.000199, 0.000140, 0.000098,
            0.000070, 0.000050, 0.000036, 0.000025, 0.000018,
            0.000013],
           [0.000705, 0.002928, 0.010482, 0.032344, 0.086011,
            0.197120, 0.389366, 0.656760, 0.972542, 1.282500,
            1.553480, 1.798500, 1.967280, 2.027300, 1.994800,
            1.900700, 1.745370, 1.554900, 1.317560, 1.030200,
            0.772125, 0.570060, 0.415254, 0.302356, 0.218502,
            0.159249, 0.112044, 0.082248, 0.060709, 0.043050,
            0.030451, 0.020584, 0.013676, 0.007918, 0.003988,
            0.001091, 0.000000, 0.000000, 0.000000, 0.000000,
            0.000000, 0.000000, 0.000000, 0.000000, 0.000000,
            0.000000, 0.000000, 0.000000, 0.000000, 0.000000,
            0.000000, 0.000000, 0.000000, 0.000000, 0.000000,
            0.000000, 0.000000, 0.000000, 0.000000, 0.000000,
            0.000000, 0.000000, 0.000000, 0.000000, 0.000000,
            0.000000, 0.000000, 0.000000, 0.000000, 0.000000,
            0.000000, 0.000000, 0.000000, 0.000000, 0.000000,
            0.000000, 0.000000, 0.000000, 0.000000, 0.000000,
            0.000000]])

    MATRIX_SRGB_D65 = np.array(
            [[3.2404542, -1.5371385, -0.4985314],
             [-0.9692660,  1.8760108,  0.0415560],
             [0.0556434, -0.2040259,  1.0572252]])

    def _gammacorr(inp):
        mask = np.zeros(inp.shape, dtype=np.float64)
        mask[inp <= 0.0031308] = 1.
        r1 = 12.92*inp
        a = 0.055
        r2 = (1 + a) * (np.maximum(inp, 0.0031308) ** (1/2.4)) - a
        return r1*mask + r2*(1.-mask)

    def lambda2xyz(lam):
        lammin = 380.
        lammax = 780.
        lam = np.asarray(lam, dtype=np.float64)
        lam = np.clip(lam, lammin, lammax)

        idx = (lam-lammin)/(lammax-lammin)*(_xyz.shape[1]-1)
        ii = np.maximum(0, np.minimum(79, int(idx)))
        w1 = 1.-(idx-ii)
        w2 = 1.-w1
        c = w1*_xyz[:, ii] + w2*_xyz[:, ii+1]
        return c

    def getxyz(n):
        E0, E1 = 1./700., 1./400.
        E = E0 + np.arange(n)*(E1-E0)/(n-1)
        res = np.zeros((3, n), dtype=np.float64)
        for i in range(n):
            res[:, i] = lambda2xyz(1./E[i])
        return res

    def to_logscale(arr, lo, hi):
        res = arr.clip(lo, hi)
        res = np.log(res/hi)
        tmp = np.log(hi/lo)
        res += tmp
        res /= tmp
        return res

    shp = spectral_cube.shape[:-1]+(3,)
    spectral_cube = spectral_cube.reshape((-1, spectral_cube.shape[-1]))
    xyz = getxyz(spectral_cube.shape[-1])
    xyz_data = np.tensordot(spectral_cube, xyz, axes=[-1, -1])
    xyz_data /= xyz_data.max()
    xyz_data = to_logscale(xyz_data, max(1e-3, xyz_data.min()), 1.)
    rgb_data = xyz_data.copy()
    for x in range(xyz_data.shape[0]):
        rgb_data[x] = _gammacorr(np.matmul(MATRIX_SRGB_D65, xyz_data[x]))
    rgb_data = rgb_data.clip(0., 1.)
    return rgb_data.reshape(shp)


def _find_closest(A, target):
    # A must be sorted
    idx = np.clip(A.searchsorted(target), 1, len(A)-1)
    idx -= target - A[idx-1] < A[idx] - target
    return idx


def _makeplot(name, block=True, dpi=None):
    import matplotlib.pyplot as plt

    if name is None:
        plt.show(block=block)
        if block:
            plt.close()
        return
    extension = os.path.splitext(name)[1]
    if extension in (".pdf", ".png", ".svg"):
        args = {}
        if dpi is not None:
            args['dpi'] = float(dpi)
        plt.savefig(name, **args)
        plt.close()
    else:
        raise ValueError("file format not understood")


def _limit_xy(**kwargs):
    import matplotlib.pyplot as plt

    x1, x2, y1, y2 = plt.axis()
    x1 = kwargs.pop("xmin", x1)
    x2 = kwargs.pop("xmax", x2)
    y1 = kwargs.pop("ymin", y1)
    y2 = kwargs.pop("ymax", y2)
    plt.axis((x1, x2, y1, y2))


def _register_cmaps():
    import matplotlib as mpl
    from matplotlib.colors import LinearSegmentedColormap

    try:
        if _register_cmaps._cmaps_registered:
            return
    except AttributeError:
        _register_cmaps._cmaps_registered = True

    planckcmap = {'red':   ((0., 0., 0.), (.4, 0., 0.), (.5, 1., 1.),
                            (.7, 1., 1.), (.8, .83, .83), (.9, .67, .67),
                            (1., .5, .5)),
                  'green': ((0., 0., 0.), (.2, 0., 0.), (.3, .3, .3),
                            (.4, .7, .7), (.5, 1., 1.), (.6, .7, .7),
                            (.7, .3, .3), (.8, 0., 0.), (1., 0., 0.)),
                  'blue':  ((0., .5, .5), (.1, .67, .67), (.2, .83, .83),
                            (.3, 1., 1.), (.5, 1., 1.), (.6, 0., 0.),
                            (1., 0., 0.))}
    he_cmap = {'red':   ((0., 0., 0.), (.167, 0., 0.), (.333, .5, .5),
                         (.5, 1., 1.), (1., 1., 1.)),
               'green': ((0., 0., 0.), (.5, 0., 0.), (.667, .5, .5),
                         (.833, 1., 1.), (1., 1., 1.)),
               'blue':  ((0., 0., 0.), (.167, 1., 1.), (.333, .5, .5),
                         (.5, 0., 0.), (1., 1., 1.))}
    fd_cmap = {'red':   ((0., .35, .35), (.1, .4, .4), (.2, .25, .25),
                         (.41, .47, .47), (.5, .8, .8), (.56, .96, .96),
                         (.59, 1., 1.), (.74, .8, .8), (.8, .8, .8),
                         (.9, .5, .5), (1., .4, .4)),
               'green': ((0., 0., 0.), (.2, 0., 0.), (.362, .88, .88),
                         (.5, 1., 1.), (.638, .88, .88), (.8, .25, .25),
                         (.9, .3, .3), (1., .2, .2)),
               'blue':  ((0., .35, .35), (.1, .4, .4), (.2, .8, .8),
                         (.26, .8, .8), (.41, 1., 1.), (.44, .96, .96),
                         (.5, .8, .8), (.59, .47, .47), (.8, 0., 0.),
                         (1., 0., 0.))}
    fdu_cmap = {'red':   ((0., 1., 1.), (0.1, .8, .8), (.2, .65, .65),
                          (.41, .6, .6), (.5, .7, .7), (.56, .96, .96),
                          (.59, 1., 1.), (.74, .8, .8), (.8, .8, .8),
                          (.9, .5, .5), (1., .4, .4)),
                'green': ((0., .9, .9), (.362, .95, .95), (.5, 1., 1.),
                          (.638, .88, .88), (.8, .25, .25), (.9, .3, .3),
                          (1., .2, .2)),
                'blue':  ((0., 1., 1.), (.1, .8, .8), (.2, 1., 1.),
                          (.41, 1., 1.), (.44, .96, .96), (.5, .7, .7),
                          (.59, .42, .42), (.8, 0., 0.), (1., 0., 0.))}
    pm_cmap = {'red':   ((0., 1., 1.), (.1, .96, .96), (.2, .84, .84),
                         (.3, .64, .64), (.4, .36, .36), (.5, 0., 0.),
                         (1., 0., 0.)),
               'green': ((0., .5, .5), (.1, .32, .32), (.2, .18, .18),
                         (.3, .8, .8),  (.4, .2, .2), (.5, 0., 0.),
                         (.6, .2, .2), (.7, .8, .8), (.8, .18, .18),
                         (.9, .32, .32), (1., .5, .5)),
               'blue':  ((0., 0., 0.), (.5, 0., 0.), (.6, .36, .36),
                         (.7, .64, .64), (.8, .84, .84), (.9, .96, .96),
                         (1., 1., 1.))}

    mpl.colormaps.register(cmap=LinearSegmentedColormap("Planck-like",
                                                        planckcmap))
    mpl.colormaps.register(cmap=LinearSegmentedColormap("High Energy",
                                                        he_cmap))
    mpl.colormaps.register(cmap=LinearSegmentedColormap("Faraday Map",
                                                        fd_cmap))
    mpl.colormaps.register(cmap=LinearSegmentedColormap("Faraday Uncertainty",
                                                        fdu_cmap))
    mpl.colormaps.register(cmap=LinearSegmentedColormap("Plus Minus", pm_cmap))


def _extract_list_kwargs(kwargs, keys, n):
    tmp = {}
    for kk in keys:
        val = kwargs.pop(kk, None)
        tmp[kk] = val if isinstance(val, list) else n*[val]
    with_legend = "label" in keys and any(ll is not None for ll in tmp["label"])
    return [{kk: vv[i] for kk, vv in tmp.items()} for i in range(n)], with_legend


def _plot_history(f, ax, **kwargs):
    import matplotlib.pyplot as plt
    from matplotlib.dates import DateFormatter, date2num

    for i, fld in enumerate(f):
        if not isinstance(fld, EnergyHistory):
            raise TypeError
    label = kwargs.pop("label", None)
    if not isinstance(label, list):
        label = [label] * len(f)
    alpha = kwargs.pop("alpha", None)
    if not isinstance(alpha, list):
        alpha = [alpha] * len(f)
    color = kwargs.pop("color", None)
    if not isinstance(color, list):
        color = [color] * len(f)
    size = kwargs.pop("s", None)
    if not isinstance(size, list):
        size = [size] * len(f)
    ax.set_title(kwargs.pop("title", ""))
    ax.set_xlabel(kwargs.pop("xlabel", ""))
    ax.set_ylabel(kwargs.pop("ylabel", ""))

    skip_timestamp_conversion = kwargs.pop("skip_timestamp_conversion", False)
    energy_differences = kwargs.pop("plot_energy_differences", False)

    plt.xscale(kwargs.pop("xscale", "linear"))
    default_yscale = 'linear' if not energy_differences else 'log'
    plt.yscale(kwargs.pop("yscale", default_yscale))

    mi, ma = np.inf, -np.inf

    for i, fld in enumerate(f):
        kwargs = {'alpha': alpha[i], 's': size[i], 'color': color[i]}

        if skip_timestamp_conversion:
            xcoord = fld.time_stamps
        else:
            xcoord = date2num([dt.fromtimestamp(ts) for ts in fld.time_stamps])

        if not energy_differences:
            ycoord = fld.energy_values
            ax.scatter(xcoord, ycoord, label=label[i], **kwargs)
        else:
            E = np.array(fld.energy_values)
            dE = E[1:] - E[:-1]
            xcoord = np.array(xcoord[1:])
            idx_pos = (dE > 0)
            idx_neg = (dE < 0)
            label_pos = label[i] + ' (pos)' if label[i] is not None else None
            label_neg = label[i] + ' (neg)' if label[i] is not None else None
            ax.scatter(xcoord[idx_pos], dE[idx_pos], marker='^',
                       label=label_pos, **kwargs)
            ax.scatter(xcoord[idx_neg], dE[idx_neg], marker='v',
                       label=label_neg, **kwargs)

        mi, ma = min([min(xcoord), mi]), max([max(xcoord), ma])

    delta = (ma-mi)*0.05
    if delta == 0.:
        delta = 1.
    ax.set_xlim((mi-delta, ma+delta))
    if not skip_timestamp_conversion:
        xfmt = DateFormatter('%H:%M')
        ax.xaxis.set_major_formatter(xfmt)
    _limit_xy(**kwargs)
    if label != ([None]*len(f)):
        plt.legend(loc="upper right")


def _plot1D(f, ax, **kwargs):
    import matplotlib.pyplot as plt

    dom = f[0].domain[0]
    add_kwargs, with_legend = _extract_list_kwargs(kwargs, ("label", "alpha", "color", "linewidth"), len(f))

    if isinstance(dom, RGSpace):
        plt.yscale(kwargs.pop("yscale", "linear"))
        npoints = dom.shape[0]
        dist = dom.distances[0]
        xcoord = np.arange(npoints, dtype=np.float64)*dist
        for i, fld in enumerate(f):
            ycoord = fld.val
            plt.plot(xcoord, ycoord, **add_kwargs[i])
    elif isinstance(dom, PowerSpace):
        plt.xscale(kwargs.pop("xscale", "log"))
        plt.yscale(kwargs.pop("yscale", "log"))
        xcoord = dom.k_lengths
        for i, fld in enumerate(f):
            ycoord = fld.val_rw()
            ycoord[0] = ycoord[1]
            plt.plot(xcoord, ycoord, **add_kwargs[i])
    else:
        raise RuntimeError("This point should never be reached")

    _limit_xy(**kwargs)
    if with_legend:
        ax.legend(loc="upper right")


def plottable1D(f):
    dom = f[0].domain
    is_1d_plottable = isinstance(dom[0], (RGSpace, PowerSpace))
    is_1d_plottable &= (len(dom) == 1) and (len(dom.shape) == 1)
    is_1d_plottable &= all(dom == el.domain for el in f)
    return is_1d_plottable


def plottable2D(fld, f_space=1):
    dom = fld.domain
    if not isinstance(dom, DomainTuple) or len(dom) > 2:
        return False
    if f_space not in [0, 1]:
        return False
    x_space = 0
    if len(dom) == 2:
        x_space = 1 - f_space
        if (not isinstance(dom[f_space], RGSpace)) or len(dom[f_space].shape) != 1:
            return False
    elif len(dom) == 0:
        return False
    if not isinstance(dom[x_space], (RGSpace, HPSpace, GLSpace)):
        return False
    if isinstance(dom[x_space], RGSpace) and not len(dom[x_space].shape) == 2:
        return False
    return True


def _plotting_args_2D(fld, f_space=1):
    from .sugar import makeField

    # check for multifrequency plotting
    have_rgb, rgb = False, None
    x_space = 0
    dom = fld.domain
    if len(dom) == 1:
        x_space = 0
    elif len(dom) == 2:
        x_space = 1 - f_space
        # Only one frequency?
        if dom[f_space].shape[0] == 1:
            fld = makeField(fld.domain[x_space], fld.val.squeeze(axis=dom.axes[f_space]))
        else:
            val = fld.val
            if f_space == 0:
                val = np.moveaxis(val, 0, -1)
            rgb = _rgb_data(val)
            have_rgb = True
    else:  # "DomainTuple can only have one or two entries.
        raise ValueError('check plottable2D before using this function')
    return fld, x_space, have_rgb, rgb


def _plot2D(f, ax, **kwargs):
    import matplotlib.pyplot as plt
    from mpl_toolkits.axes_grid1 import make_axes_locatable

    f = f[0]
    dom = f.domain

    f, x_space, have_rgb, rgb = _plotting_args_2D(f, kwargs.pop("freq_space_idx", 1))

    foo = kwargs.pop("norm", None)
    norm = {} if foo is None else {'norm': foo}

    foo = kwargs.pop("aspect", None)
    aspect = {} if foo is None else {'aspect': foo}

    dom = dom[x_space]
    if not have_rgb:
        cmap = kwargs.pop("cmap", plt.rcParams['image.cmap'])

    if isinstance(dom, RGSpace):
        nx, ny = dom.shape
        dx, dy = dom.distances
        if have_rgb:
            im = ax.imshow(
                rgb, extent=[0, nx*dx, 0, ny*dy], origin="lower", **norm,
                **aspect)
        else:
            im = ax.imshow(
                f.val.T, extent=[0, nx*dx, 0, ny*dy],
                vmin=kwargs.get("vmin"), vmax=kwargs.get("vmax"),
                cmap=cmap, origin="lower", **norm, **aspect)
            cax = make_axes_locatable(ax).append_axes("right", size="5%", pad=0.05)
            plt.colorbar(im, cax=cax)
        _limit_xy(**kwargs)
        return
    elif isinstance(dom, (HPSpace, GLSpace)):
        from ducc0.healpix import Healpix_Base
        from ducc0.misc import GL_thetas

        xsize = 800
        res, mask, theta, phi = _mollweide_helper(xsize)
        if have_rgb:
            res = np.full(shape=res.shape+(3,), fill_value=1.,
                          dtype=np.float64)

        if isinstance(dom, HPSpace):
            ptg = np.empty((phi.size, 2), dtype=np.float64)
            ptg[:, 0] = theta
            ptg[:, 1] = phi
            base = Healpix_Base(int(np.sqrt(dom.size//12)), "RING")
            if have_rgb:
                res[mask] = rgb[base.ang2pix(ptg)]
            else:
                res[mask] = f.val[base.ang2pix(ptg)]
        else:
            ra = np.linspace(0, 2*np.pi, dom.nlon+1)
            dec = GL_thetas(dom.nlat)
            ilat = _find_closest(dec, theta)
            ilon = _find_closest(ra, phi)
            ilon = np.where(ilon == dom.nlon, 0, ilon)
            if have_rgb:
                res[mask] = rgb[ilat*dom[0].nlon + ilon]
            else:
                res[mask] = f.val[ilat*dom.nlon + ilon]
        plt.axis('off')
        if have_rgb:
            plt.imshow(res, origin="lower")
        else:
            plt.imshow(res, vmin=kwargs.get("vmin"), vmax=kwargs.get("vmax"),
                       norm=norm.get('norm'), cmap=cmap, origin="lower")
            plt.colorbar(orientation="horizontal")
        return
    raise ValueError("Field type not(yet) supported")


def _plotHist(f, ax, **kwargs):
    add_kwargs, with_legend = _extract_list_kwargs(kwargs, ("label", "alpha", "color", "range"),
                                                   len(f))
    add_kwargs2 = {
        "log": kwargs.pop("log", False),
        "density": kwargs.pop("density", False),
        "bins": kwargs.pop("bins", 50)
    }
    for i, fld in enumerate(f):
        ax.hist(fld.val.ravel(), **add_kwargs[i], **add_kwargs2)
    if with_legend:
        ax.legend(loc="upper right")


def _plot(f, ax, **kwargs):
    _register_cmaps()
    if isinstance(f[0], EnergyHistory):
        _plot_history(f, ax, **kwargs)
        return

    ax.set_title(kwargs.pop("title", ""))
    ax.set_xlabel(kwargs.pop("xlabel", ""))
    ax.set_ylabel(kwargs.pop("ylabel", ""))

    if plottable1D(f):
        _plot1D(f, ax, **kwargs)
        return
    if plottable2D(f[0], kwargs.get("freq_space_idx", 1)):
        _plot2D(f, ax, **kwargs)
        return
    _plotHist(f, ax, **kwargs)


class Plot:
    def __init__(self):
        self._plots = []
        self._kwargs = []

    def add(self, f, **kwargs):
        """Add a figure to the current list of plots.

        Notes
        -----
        After doing one or more calls `add()`, one needs to call `output()` to
        show or save the plot.

        Parameters
        ----------
        f : :class:`nifty8.field.Field` or list of :class:`nifty8.field.Field` or None
            If `f` is a single Field, it must be defined on a single `RGSpace`,
            `PowerSpace`, `HPSpace`, `GLSpace`.
            If it is a list, all list members must be Fields defined over the
            same one-dimensional `RGSpace` or `PowerSpace`.
            If `f` is `None`, an empty panel will be displayed.

        Optional Parameters
        -------------------
        title: string
            Title of the plot.
        xlabel: string
            Label for the x axis.
        ylabel: string
            Label for the y axis.
        [xyz]min, [xyz]max: float
            Limits for the values to plot.
        cmap: string
            Color map to use for the plot (if it is a 2D plot).
        linewidth: float or list of floats
            Line width.
        label: string of list of strings
            Annotation string.
        alpha: float or list of floats
            Transparency value.
        freq_space_idx: int
            for multi-frequency plotting: index of frequency space in domain
        """
        if f is None:
            self._plots.append(None)
            self._kwargs.append({})
            return
        if isinstance(f, (MultiField, Field, EnergyHistory)):
            f = [f]
        if not isinstance(f[0], (MultiField, Field, EnergyHistory)):
            raise TypeError("Incorrect data type. You can only add Fields or EnergyHistories, or None")
        if hasattr(f, "__len__") and all(isinstance(ff, MultiField) for ff in f):
            for kk in f[0].domain.keys():
                self._plots.append([ff[kk] for ff in f])
                mykwargs = kwargs.copy()
                if 'title' in kwargs:
                    mykwargs['title'] = "{} {}".format(kk, kwargs['title'])
                else:
                    mykwargs['title'] = "{}".format(kk)
                self._kwargs.append(mykwargs)
            return

        if isinstance(f[0], EnergyHistory):
            dom = None
        else:
            dom = f[0].domain

        if isinstance(dom, DomainTuple) \
                and any(isinstance(dd, RGSpace) for dd in dom) \
                and not "freq_space_idx" in kwargs \
                and not plottable2D(f[0], kwargs.get("freq_space_idx", 1)):
            from .sugar import makeField

            dims = [len(dd.shape) for dd in dom]
            # One space is 2d, the rest is 1d
            if np.sum(np.array(dims) == 2) == 1 and np.sum(np.array(dims) == 1) == len(dims) - 1:
                twod_index = dims.index(2)
                sizes = [dd.size for dd in dom]
                del(sizes[twod_index])
                for multi_index in product(*[tuple(range(ii)) for ii in sizes]):
                    multi_index = list(multi_index)
                    multi_index.insert(twod_index, slice(None))
                    multi_index.insert(twod_index, slice(None))
                    for ifield in range(len(f)):
                        arr = f[ifield].val[tuple(multi_index)]
                        myassert(arr.ndim == 2)
                        self._plots.append([makeField(dom[twod_index], arr)])
                        self._kwargs.append(kwargs)
                return
        self._plots.append(f)
        self._kwargs.append(kwargs)

    def output(self, **kwargs):
        """Plot the accumulated list of figures.

        Parameters
        ----------
        title: string
            Title of the full plot.
        nx, ny: int
            Number of subplots to use in x- and y-direction.
            Default: square root of the numer of plots, rounded up.
        xsize, ysize: float
            Dimensions of the full plot in inches. Default: 6.
        name: string
            If left empty, the plot will be shown on the screen,
            otherwise it will be written to a file with the given name.
            Supported extensions: .png and .pdf. Default: None.
        block: bool
            Override the blocking behavior of the non-interactive plotting
            mode. The plot will not be closed in this case but is left open!
        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            warn("Since matplotlib is not installed, NIFTy will not generate any plots.")
            return

        nplot = len(self._plots)
        if nplot == 0:
            raise ValueError("Use .add to add plots to your plotting routine.")
        fig = plt.figure()
        if "title" in kwargs:
            plt.suptitle(kwargs.pop("title"))
        nx = kwargs.pop("nx", 0)
        ny = kwargs.pop("ny", 0)
        if nx == ny == 0:
            ny = int(np.ceil(np.sqrt(nplot)))
            nx = int(np.ceil(nplot/ny))
            myassert(nx*ny >= nplot)
        elif nx == 0:
            nx = int(np.ceil(nplot/ny))
        elif ny == 0:
            ny = int(np.ceil(nplot/nx))
        if nx*ny < nplot:
            raise ValueError(
                'Figure dimensions not sufficient for number of plots. '
                'Available plot slots: {}, number of plots: {}'
                .format(nx*ny, nplot))
        xsize = kwargs.pop("xsize", 6*nx)
        ysize = kwargs.pop("ysize", 6*ny)
        fig.set_size_inches(xsize, ysize)
        for i in range(nplot):
            if self._plots[i] is None:
                continue
            ax = fig.add_subplot(ny, nx, i+1)
            _plot(self._plots[i], ax, **self._kwargs[i])
        fig.tight_layout()
        _makeplot(kwargs.pop("name", None),
                  block=kwargs.pop("block", True),
                  dpi=kwargs.pop("dpi", None))



from .version import __version__

from . import config
from . import random

from .domains import *

from .domain_tuple import DomainTuple
from .multi_domain import MultiDomain
from .field import Field
from .multi_field import MultiField


from .operators.operator import Operator
from .operators.linear_operator import LinearOperator

from .operators.adder import Adder
from .operators.diagonal_operator import DiagonalOperator
from .operators.distributors import DOFDistributor, PowerDistributor
from .operators.domain_tuple_field_inserter import DomainTupleFieldInserter
from .operators.einsum import LinearEinsum, MultiLinearEinsum
from .operators.contraction_operator import ContractionOperator, IntegrationOperator
from .operators.linear_interpolation import LinearInterpolator
from .operators.endomorphic_operator import EndomorphicOperator
from .operators.harmonic_operators import (
    FFTOperator, HartleyOperator, SHTOperator, HarmonicTransformOperator,
    HarmonicSmoothingOperator)
from .operators.field_zero_padder import FieldZeroPadder
from .operators.inversion_enabler import InversionEnabler
from .operators.mask_operator import MaskOperator
from .operators.regridding_operator import RegriddingOperator
from .operators.sampling_enabler import SamplingEnabler
from .operators.sandwich_operator import SandwichOperator
from .operators.scaling_operator import ScalingOperator
from .operators.selection_operators import SliceOperator, SplitOperator
from .operators.block_diagonal_operator import BlockDiagonalOperator
from .operators.outer_product_operator import OuterProduct
from .operators.simple_linear_operators import (
    VdotOperator, ConjugationOperator, Realizer, FieldAdapter, ducktape,
    GeometryRemover, NullOperator, PartialExtractor, Imaginizer, PrependKey,
    DomainChangerAndReshaper, ExtractAtIndices)
from .operators.matrix_product_operator import MatrixProductOperator
from .operators.value_inserter import ValueInserter
from .operators.energy_operators import (
    EnergyOperator, GaussianEnergy, PoissonianEnergy, InverseGammaEnergy,
    BernoulliEnergy, StandardHamiltonian, AveragedEnergy, QuadraticFormOperator,
    Squared2NormOperator, StudentTEnergy, VariableCovarianceGaussianEnergy,
    LikelihoodEnergyOperator)
from .operators.convolution_operators import FuncConvolutionOperator
from .operators.normal_operators import NormalTransform, LognormalTransform
from .operators.multifield2vector import Multifield2Vector
from .operators.jax_operator import *
from .operators.counting_operator import CountingOperator
from .operators.transpose_operator import TransposeOperator

from .probing import probe_with_posterior_samples, probe_diagonal, \
    StatCalculator, approximation2endo

from .minimization.line_search import LineSearch
from .minimization.iteration_controllers import (
    IterationController, GradientNormController, DeltaEnergyController,
    GradInfNormController, AbsDeltaEnergyController, StochasticAbsDeltaEnergyController)
from .minimization.minimizer import Minimizer
from .minimization.conjugate_gradient import ConjugateGradient
from .minimization.nonlinear_cg import NonlinearCG
from .minimization.descent_minimizers import (
    DescentMinimizer, SteepestDescent, VL_BFGS, L_BFGS, RelaxedNewton,
    NewtonCG)
from .minimization.stochastic_minimizer import ADVIOptimizer
from .minimization.scipy_minimizer import L_BFGS_B
from .minimization.energy import Energy
from .minimization.quadratic_energy import QuadraticEnergy
from .minimization.sample_list import SampleList, SampleListBase, ResidualSampleList
from .minimization.energy_adapter import EnergyAdapter, StochasticEnergyAdapter
from .minimization.kl_energies import SampledKLEnergy, SampledKLEnergyClass
from .minimization.optimize_kl import optimize_kl
from .minimization.config.optimize_kl_config import OptimizeKLConfig

from .sugar import *

from .plot import Plot

from .library.special_distributions import InverseGammaOperator, \
    UniformOperator, LaplaceOperator, LogInverseGammaOperator, \
    GammaOperator
from .library.los_response import LOSResponse
from .library.dynamic_operator import (dynamic_operator,
                                       dynamic_lightcone_operator)
from .library.light_cone_operator import LightConeOperator

from .library.wiener_filter_curvature import WienerFilterCurvature
from .library.adjust_variances import (make_adjust_variances_hamiltonian,
                                       do_adjust_variances)
from .library.nft import Gridder, FinuFFT, Nufft
from .library.correlated_fields import CorrelatedFieldMaker
from .library.correlated_fields_simple import SimpleCorrelatedField
from .library.variational_models import MeanFieldVI, FullCovarianceVI

from . import extra

from .utilities import memo, frozendict, myassert

from .logger import logger

from .linearization import Linearization

from .operator_spectrum import operator_spectrum
from .evidence_lower_bound import estimate_evidence_lower_bound

from .operator_tree_optimiser import optimise_operator

from .ducc_dispatch import set_nthreads, nthreads

try:
    from . import re
    from . import nifty2jax
except ImportError:
    pass

# We deliberately don't set __all__ here, because we don't want people to do a
# "from nifty8 import *"; that would swamp the global namespace.



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
# Authors: Gordian Edenhofer, Philipp Frank
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
from nifty8.extra import check_linear_operator, check_operator
from numpy.testing import assert_allclose

from ..common import list2fixture, setup_function, teardown_function

spaces = (ift.UnstructuredDomain(4),
          ift.RGSpace((3, 2)),
          ift.LMSpace(5),
          ift.GLSpace(4))

space1 = list2fixture(spaces)
space2 = list2fixture(spaces)
dtype = list2fixture([np.float64, np.complex128])


def test_linear_einsum_outer(space1, space2, dtype, n_invocations=10):
    mf_dom = ift.MultiDomain.make({
        "dom01": space1,
        "dom02": ift.DomainTuple.make((space1, space2))})
    mf = ift.from_random(mf_dom, "normal", dtype=dtype)
    ss = "i,ij,j->ij"
    key_order = ("dom01", "dom02")
    le = ift.LinearEinsum(space2, mf, ss, key_order=key_order)
    ift.myassert(check_linear_operator(le, domain_dtype=dtype, target_dtype=dtype) is None)

    le_ift = ift.DiagonalOperator(mf["dom01"], domain=mf_dom["dom02"], spaces=0) @ ift.DiagonalOperator(mf["dom02"])
    le_ift = le_ift @ ift.OuterProduct(ift.DomainTuple.make(mf_dom["dom02"][1]),
                                       ift.full(mf_dom["dom01"], 1.))

    for _ in range(n_invocations):
        r = ift.from_random(le.domain, "normal", dtype=dtype)
        assert_allclose(le(r).val, le_ift(r).val)
        r_adj = ift.from_random(le.target, "normal", dtype=dtype)
        assert_allclose(le.adjoint(r_adj).val, le_ift.adjoint(r_adj).val)


def test_linear_einsum_contraction(space1, space2, dtype, n_invocations=10):
    mf_dom = ift.MultiDomain.make({
        "dom01": space1,
        "dom02": ift.DomainTuple.make((space1, space2))})
    mf = ift.from_random(mf_dom, "normal", dtype=dtype)
    ss = "i,ij,j->i"
    key_order = ("dom01", "dom02")
    le = ift.LinearEinsum(space2, mf, ss, key_order=key_order)
    ift.myassert(check_linear_operator(le, domain_dtype=dtype, target_dtype=dtype) is None)

    le_ift = ift.ContractionOperator(mf_dom["dom02"], 1)
    le_ift = le_ift @ ift.DiagonalOperator(mf["dom01"], domain=mf_dom["dom02"], spaces=0)
    le_ift = le_ift @ ift.DiagonalOperator(mf["dom02"])
    le_ift = le_ift @ ift.OuterProduct(ift.DomainTuple.make(mf_dom["dom02"][1]),
                                       ift.full(mf_dom["dom01"], 1.),)

    for _ in range(n_invocations):
        r = ift.from_random(le.domain, "normal", dtype=dtype)
        assert_allclose(le(r).val, le_ift(r).val)
        r_adj = ift.from_random(le.target, "normal", dtype=dtype)
        assert_allclose(le.adjoint(r_adj).val, le_ift.adjoint(r_adj).val)


class _SwitchSpacesOperator(ift.LinearOperator):
    """Operator to permutate the domain entries of fields.

    Exchanges the entries `space1` and `space2` of the input's domain.
    """
    def __init__(self, domain, space1, space2=0):
        self._capability = self.TIMES | self.ADJOINT_TIMES
        self._domain = ift.DomainTuple.make(domain)

        n_spaces = len(self._domain)
        if space1 >= n_spaces or space1 < 0 \
           or space2 >= n_spaces or space2 < 0:
            raise ValueError("invalid space value")

        tgt = list(self._domain)
        tgt[space2] = self._domain[space1]
        tgt[space1] = self._domain[space2]
        self._target = ift.DomainTuple.make(tgt)

        dom_axes = self._domain.axes
        tgt_axes = self._target.axes

        self._axes_dom = dom_axes[space1] + dom_axes[space2]
        self._axes_tgt = tgt_axes[space2] + tgt_axes[space1]

    def apply(self, x, mode):
        self._check_input(x, mode)
        args = self._axes_dom, self._axes_tgt
        if mode == self.ADJOINT_TIMES:
            args = args[::-1]
        return ift.Field(self._tgt(mode), np.moveaxis(x.val, *args))


def test_linear_einsum_transpose(space1, space2, dtype, n_invocations=10):
    dom = ift.DomainTuple.make((space1, space2))
    mf = ift.MultiField.from_dict({})
    ss = "ij->ji"
    le = ift.LinearEinsum(dom, mf, ss)
    ift.myassert(check_linear_operator(le, domain_dtype=dtype, target_dtype=dtype) is None)

    # SwitchSpacesOperator is equivalent to LinearEinsum with "ij->ji"
    le_ift = _SwitchSpacesOperator(dom, 1)

    for _ in range(n_invocations):
        r = ift.from_random(le.domain, "normal", dtype=dtype)
        assert_allclose(le(r).val, le_ift(r).val)
        r_adj = ift.from_random(le.target, "normal", dtype=dtype)
        assert_allclose(le.adjoint(r_adj).val, le_ift.adjoint(r_adj).val)


def test_multi_linear_einsum_outer(space1, space2, dtype):
    ntries = 10
    n_invocations = 5
    mf_dom = ift.MultiDomain.make({
        "dom01": space1,
        "dom02": ift.DomainTuple.make((space1, space2)),
        "dom03": space2})
    ss = "i,ij,j->ij"
    key_order = ("dom01", "dom02", "dom03")
    mle = ift.MultiLinearEinsum(mf_dom, ss, key_order=key_order)
    check_operator(mle, ift.from_random(mle.domain, "normal", dtype=dtype), ntries=ntries)

    outer_i = ift.OuterProduct(
        ift.DomainTuple.make(mf_dom["dom02"][0]), ift.full(mf_dom["dom03"], 1.)
    )
    outer_j = ift.OuterProduct(
        ift.DomainTuple.make(mf_dom["dom02"][1]), ift.full(mf_dom["dom01"], 1.)
    )
    # SwitchSpacesOperator is equivalent to LinearEinsum with "ij->ji"
    mle_ift = _SwitchSpacesOperator(outer_i.target, 1) @ outer_i @ \
        ift.FieldAdapter(mf_dom["dom01"], "dom01") * \
        ift.FieldAdapter(mf_dom["dom02"], "dom02") * \
        (outer_j @ ift.FieldAdapter(mf_dom["dom03"], "dom03"))

    for _ in range(n_invocations):
        rl = ift.Linearization.make_var(ift.from_random(mle.domain, "normal", dtype=dtype))
        mle_rl, mle_ift_rl = mle(rl), mle_ift(rl)
        assert_allclose(mle_rl.val.val, mle_ift_rl.val.val)
        assert_allclose(mle_rl.jac(rl.val).val, mle_ift_rl.jac(rl.val).val)

        rj_adj = ift.from_random(mle_rl.jac.target, "normal", dtype=dtype)
        mle_j_val = mle_rl.jac.adjoint(rj_adj).val
        mle_ift_j_val = mle_ift_rl.jac.adjoint(rj_adj).val
        for k in mle_ift.domain.keys():
            assert_allclose(mle_j_val[k], mle_ift_j_val[k])



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from nifty8.library.variational_models import GaussianEntropy
from numpy.testing import assert_allclose

from ..common import list2fixture, setup_function, teardown_function

pmp = pytest.mark.parametrize


@pmp("N", [17, 32])
def test_gaussian_entropy(N):
    linear_space = ift.RGSpace(N)
    fld = ift.from_random(linear_space, 'uniform')
    # minus due to subtraction in KL
    entropy = -0.5*np.sum(np.log(2*fld.val*fld.val*np.pi*np.e))
    myEntropy = GaussianEntropy(fld.domain)
    assert_allclose(entropy, myEntropy(fld).val)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from ..utilities import NiftyMeta


class Energy(metaclass=NiftyMeta):
    """Provides the functional used by minimization schemes.

    The Energy object is an implementation of a scalar function including its
    gradient and metric at some position.

    Parameters
    ----------
    position : :class:`nifty8.field.Field`
        The input parameter of the scalar function.

    Notes
    -----
    An instance of the Energy class is defined at a certain location. If one
    is interested in the value, gradient or metric of the abstract energy
    functional one has to 'jump' to the new position using the `at` method.
    This method returns a new energy instance residing at the new position. By
    this approach, intermediate results from computing e.g. the gradient can
    safely be reused for e.g. the value or the metric.

    Memorizing the evaluations of some quantities minimizes the computational
    effort for multiple calls.
    """

    def __init__(self, position):
        self._position = position
        self._gradnorm = None

    def at(self, position):
        """Returns a new Energy object, initialized at `position`.

        Parameters
        ----------
        position : :class:`nifty8.field.Field`
            Location in parameter space for the new Energy object.

        Returns
        -------
        Energy
            Energy object at new position.
        """
        return self.__class__(position)

    @property
    def position(self):
        """
        field : selected location in parameter space.

        The Field location in parameter space where value, gradient and
        metric are evaluated.
        """
        return self._position

    @property
    def value(self):
        """
        float : value of the functional.

            The value of the energy functional at given `position`.
        """
        raise NotImplementedError

    @property
    def gradient(self):
        """
        field : The gradient at given `position`.
        """
        raise NotImplementedError

    @property
    def gradient_norm(self):
        """
        float : L2-norm of the gradient at given `position`.
        """
        if self._gradnorm is None:
            self._gradnorm = self.gradient.norm()
        return self._gradnorm

    @property
    def metric(self):
        """
        LinearOperator : implicitly defined metric.
            A positive semi-definite operator or function describing the
            metric of the potential at the given `position`.
        """
        raise NotImplementedError

    def apply_metric(self, x):
        """
        Parameters
        ----------
        x : :class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField`
            Argument for the metric operator

        Returns
        -------
        :class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField`
            Output of the metric operator
        """
        raise NotImplementedError

    def longest_step(self, direction):
        """Returns the longest allowed step size along `direction`

        Parameters
        ----------
        direction : :class:`nifty8.field.Field`
            the search direction

        Returns
        -------
        float or None
            the longest allowed step when starting from `self.position` along
            `direction`. If None, the step size is not limited.
        """
        return None



#!/usr/bin/env python3

# Copyright(C) 2013-2021 Max-Planck-Society
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

# %%
#
# WARNING: This code does not behave deterministically. It works fine when
# executing cell by cell using vscodes notebook functionality but when running
# from the command line with either python3 or ipython3 the following happens:
# This is probably due to an issue with host_callback.
# Concretely it just stops adding points to the debug list after some random
# number of leapfrog steps.
#

import jax.numpy as jnp
import matplotlib
import matplotlib.pyplot as plt

import nifty8.re as jft

# %%
jft.hmc._DEBUG_FLAG = True

# %%
cov = jnp.array([10., 1.])

potential_energy = lambda q: jnp.sum(0.5 * q**2 / cov)

initial_position = jnp.array([1., 1.])

sampler = jft.NUTSChain(
    potential_energy=potential_energy,
    inverse_mass_matrix=1.,
    position_proto=initial_position,
    step_size=0.12,
    max_tree_depth=10,
)

# %%
jft.hmc._DEBUG_STORE = []
jft.hmc._DEBUG_TREE_END_IDXS = []
jft.hmc._DEBUG_SUBTREE_END_IDXS = []

chain, _ = sampler.generate_n_samples(
    48, initial_position, num_samples=5, save_intermediates=True
)

plt.hist(chain.depths)
plt.show()

# %%
debug_pos = jnp.array([qp.position for qp in jft.hmc._DEBUG_STORE])
print(len(debug_pos))

# %%
prop_cycle = plt.rcParams['axes.prop_cycle']
colors = prop_cycle.by_key()['color']

ax = plt.gca()
ellipse = matplotlib.patches.Ellipse(
    xy=(0, 0),
    width=jnp.sqrt(cov[0]),
    height=jnp.sqrt(cov[1]),
    edgecolor='k',
    fc='None',
    lw=1
)
ax.add_patch(ellipse)

color_idx = 0
start_and_end_idxs = zip(
    [
        0,
    ] + jft.hmc._DEBUG_SUBTREE_END_IDXS[:-1], jft.hmc._DEBUG_SUBTREE_END_IDXS
)
for start_idx, end_idx in start_and_end_idxs:
    slice = debug_pos[start_idx:end_idx]
    ax.plot(
        slice[:, 0],
        slice[:, 1],
        '-o',
        markersize=1,
        linewidth=0.5,
        color=colors[color_idx % len(colors)]
    )
    if end_idx in jft.hmc._DEBUG_TREE_END_IDXS:
        color_idx = (color_idx + 1) % len(colors)

ax.scatter(
    chain.samples[:, 0],
    chain.samples[:, 1],
    marker='x',
    color='k',
    label='samples'
)
ax.scatter(initial_position[0], initial_position[1], label='starting position')
ax.set_xlabel('x')
ax.set_ylabel('y')
ax.legend()

fig_width_pt = 426  # pt (a4paper, and such)
# fig_width_pt = 360 # pt
inches_per_pt = 1 / 72.27
fig_width_in = 0.9 * fig_width_pt * inches_per_pt
fig_height_in = fig_width_in * 0.618
fig_dims = (fig_width_in, fig_height_in)

plt.tight_layout()
plt.show()
plt.savefig("trajectories.pdf", bbox_inches='tight')



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import pytest
from nifty8 import DomainChangerAndReshaper

from ..common import list2fixture, setup_function, teardown_function


def test_reshaper():
    dom = ift.makeDomain((ift.RGSpace([2, 4]), ift.UnstructuredDomain([10])))
    fld = ift.full(dom, 0.)
    lin = ift.Linearization.make_var(fld)
    for tgt in [ift.UnstructuredDomain(dom.size), ift.RGSpace([8, 10]), (ift.UnstructuredDomain(4), ift.RGSpace(20))]:
        tgt = ift.makeDomain(tgt)
        op = DomainChangerAndReshaper(dom, tgt)
        ift.extra.check_linear_operator(op)

        op1 = ift.ScalingOperator(dom, 1.2)

        with pytest.raises(ValueError):
            op1 @ op

        op2 = op1 @ op.ducktape_left(op1.domain)
        op3 = op1.ducktape(op.target) @ op
        ift.extra.check_linear_operator(op2)
        ift.extra.check_linear_operator(op3)

        assert fld.ducktape_left(tgt).domain is tgt
        assert lin.ducktape_left(tgt).target is tgt



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society
# Author: Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from tempfile import TemporaryDirectory

import nifty8 as ift
import numpy as np
import pytest

from ..common import setup_function, teardown_function

comm = ift.utilities.get_MPI_params()[0]
master = True if comm is None else comm.Get_rank() == 0

pmp = pytest.mark.parametrize

point_estimates = []

dom = ift.RGSpace([20, 20])
logsky = ift.SimpleCorrelatedField(dom, 0., None, (1, 1), None, None, (-3, 1))
sky = logsky.exp()
mock_pos = ift.from_random(logsky.domain)


def cstfunc(iglobal):
    if iglobal == 0:
        return []
    if iglobal == 1:
        return ["xi"]
    if iglobal in [2, 3]:
        return ["fluctuations"]
    if iglobal == 4:
        return []
    return []


def pefunc(iglobal):
    if iglobal == 0:
        return []
    if iglobal == 1:
        return []
    if iglobal == 2:
        return ["xi"]
    if iglobal == 3:
        return ["fluctuations"]
    if iglobal == 4:
        return ["xi"]
    return []


ic = ift.GradientNormController(iteration_limit=5)

@pmp("constants", [[], ["fluctuations"], cstfunc])
@pmp("point_estimates", [[], ["fluctuations"], pefunc])
@pmp("kl_minimizer", [ift.SteepestDescent(ic),
                      lambda n: ift.NewtonCG(ic) if n < 2 else ift.VL_BFGS(ic)])
@pmp("sampling_iteration_controller", [ic])
@pmp("nonlinear_sampling_minimizer", [None, ift.NewtonCG(ift.GradInfNormController(1e-3, iteration_limit=1))])
@pmp("n_samples", [0, 2])
@pmp("dry_run", [False, True])
def test_optimize_kl(constants, point_estimates, kl_minimizer, n_samples,
                     sampling_iteration_controller, nonlinear_sampling_minimizer,
                     dry_run):
    final_index = 5

    foo, output_directory = _create_temp_outputdir()
    bar, output_directory1 = _create_temp_outputdir()

    initial_position = None
    initial_index = 0

    d = sky(mock_pos)
    likelihood_energy = ift.GaussianEnergy(d) @ sky
    inspect_callback = None
    terminate_callback = None
    transitions = None
    export_operator_outputs = {}
    rand_state = ift.random.getState()
    sl = ift.optimize_kl(likelihood_energy, final_index, n_samples, kl_minimizer,
                         sampling_iteration_controller, nonlinear_sampling_minimizer, constants,
                         point_estimates, transitions, export_operator_outputs, output_directory, initial_position,
                         initial_index, comm, inspect_callback,
                         terminate_callback, save_strategy="all", plot_energy_history=False,
                         plot_minisanity_history=False, dry_run=dry_run)

    ift.random.setState(rand_state)

    def terminate_callback(iglobal):
        return iglobal in [0, 3]

    for _ in range(5):
        sl1 = ift.optimize_kl(likelihood_energy, final_index, n_samples, kl_minimizer,
                              sampling_iteration_controller, nonlinear_sampling_minimizer, constants,
                              point_estimates, transitions, export_operator_outputs, output_directory1, initial_position,
                              initial_index, comm, inspect_callback,
                              terminate_callback, resume=True, save_strategy="last",
                              plot_energy_history=False, plot_minisanity_history=False, dry_run=dry_run)

    for aa, bb in zip(sl.iterator(), sl1.iterator()):
        ift.extra.assert_allclose(aa, bb)


def test_transitions():
    final_index = 3

    foo, output_directory = _create_temp_outputdir()

    initial_position = None
    initial_index = 0
    n_samples = 2

    kl_minimizer = ift.SteepestDescent(ic)
    # ift.NewtonCG(ift.GradInfNormController(1e-3, iteration_limit=1))
    sampling_iteration_controller = ic
    nonlinear_sampling_minimizer = None
    constants = []
    point_estimates = []

    dom0 = ift.UnstructuredDomain(5)
    dom1 = ift.UnstructuredDomain(7)
    mdom0 = ift.makeDomain({"in0": dom0})
    mdom1 = ift.makeDomain({"in0": dom1})
    mdom2 = ift.makeDomain({"in1": dom1})


    def likelihood_energy(iglobal):
        if iglobal == 0:
            return ift.GaussianEnergy(domain=mdom0, sampling_dtype=float)
        elif iglobal == 1:
            return ift.GaussianEnergy(domain=mdom1, sampling_dtype=float)
        else:
            return ift.GaussianEnergy(domain=mdom2, sampling_dtype=float)


    def transitions(iglobal):
        if iglobal == 0:
            return None
        elif iglobal == 1:
            mask = np.zeros(dom1.shape)
            mask[:2] = 1
            op = ift.MaskOperator(ift.makeField(dom1, mask)).adjoint.ducktape("in0").ducktape_left("in0")
            return lambda sl: op(sl._m)
        elif iglobal == 2:
            fa = ift.FieldAdapter(dom1, "in0")
            op = fa.ducktape_left("in1")
            return lambda sl: op(sl._m)

        raise RuntimeError


    inspect_callback = None
    terminate_callback = None
    export_operator_outputs = {}

    sl = ift.optimize_kl(likelihood_energy, final_index, n_samples, kl_minimizer,
                         sampling_iteration_controller, nonlinear_sampling_minimizer, constants,
                         point_estimates, transitions, export_operator_outputs, output_directory, initial_position,
                         initial_index, comm, inspect_callback, terminate_callback, save_strategy="all",
                         plot_energy_history=False, plot_minisanity_history=False)
    assert sl.domain is mdom2


def _create_temp_outputdir():
    if master:
        output_directory = TemporaryDirectory()
        name = output_directory.name
    else:
        name = output_directory = None
    if comm is not None:
        name = comm.bcast(name, 0)
    return output_directory, name



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
from numpy.testing import assert_allclose, assert_equal

from ..common import list2fixture, setup_function, teardown_function

space1 = list2fixture([
    ift.RGSpace(4),
    ift.PowerSpace(ift.RGSpace((4, 4), harmonic=True)),
    ift.LMSpace(5),
    ift.HPSpace(4),
    ift.GLSpace(4)
])
space2 = space1


def test_times_adjoint_times(space1, space2):
    cspace = (space1, space2)
    diag1 = ift.Field.from_random(domain=space1, random_type='normal')
    diag2 = ift.Field.from_random(domain=space2, random_type='normal')
    op1 = ift.DiagonalOperator(diag1, cspace, spaces=(0,))
    op2 = ift.DiagonalOperator(diag2, cspace, spaces=(1,))

    op = op2(op1)

    rand1 = ift.Field.from_random(domain=(space1, space2), random_type='normal')
    rand2 = ift.Field.from_random(domain=(space1, space2), random_type='normal')

    tt1 = rand2.s_vdot(op.times(rand1))
    tt2 = rand1.s_vdot(op.adjoint_times(rand2))
    assert_allclose(tt1, tt2)


def test_times_inverse_times(space1, space2):
    cspace = (space1, space2)
    diag1 = ift.Field.from_random(domain=space1, random_type='normal')
    diag2 = ift.Field.from_random(domain=space2, random_type='normal')
    op1 = ift.DiagonalOperator(diag1, cspace, spaces=(0,))
    op2 = ift.DiagonalOperator(diag2, cspace, spaces=(1,))

    op = op2(op1)

    rand1 = ift.Field.from_random(domain=(space1, space2), random_type='normal')
    tt1 = op.inverse_times(op.times(rand1))

    assert_allclose(tt1.val, rand1.val)


def test_sum(space1):
    op1 = ift.makeOp(ift.Field.full(space1, 2.))
    op2 = ift.ScalingOperator(space1, 3.)
    full_op = op1 + op2 - (op2 - op1) + op1 + op1 + op2
    x = ift.Field.full(space1, 1.)
    res = full_op(x)
    assert_equal(isinstance(full_op, ift.DiagonalOperator), True)
    assert_allclose(res.val, 11.)


def test_chain(space1):
    op1 = ift.makeOp(ift.Field.full(space1, 2.))
    op2 = ift.ScalingOperator(space1, 3.,)
    full_op = op1(op2)(op2)(op1)(op1)(op1)(op2)
    x = ift.Field.full(space1, 1.)
    res = full_op(x)
    assert_equal(isinstance(full_op, ift.DiagonalOperator), True)
    assert_allclose(res.val, 432.)


def test_mix(space1):
    op1 = ift.makeOp(ift.Field.full(space1, 2.))
    op2 = ift.ScalingOperator(space1, 3.)
    full_op = op1(op2 + op2)(op1)(op1) - op1(op2)
    x = ift.Field.full(space1, 1.)
    res = full_op(x)
    assert_equal(isinstance(full_op, ift.DiagonalOperator), True)
    assert_allclose(res.val, 42.)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
from nifty8.operators.simplify_for_const import ConstantOperator
from numpy.testing import assert_allclose, assert_raises


def test_simplification():
    dom = {"a": ift.RGSpace(10)}
    f1 = ift.full(dom, 2.)
    op = ift.FFTOperator(f1.domain["a"]).ducktape("a")
    _, op2 = op.simplify_for_constant_input(f1)
    ift.myassert(isinstance(op2, ConstantOperator))
    assert_allclose(op(f1).val, op2.force(f1).val)

    dom = {"a": ift.RGSpace(10), "b": ift.RGSpace(5)}
    f1 = ift.full(dom, 2.)
    pdom = {"a": ift.RGSpace(10)}
    f2 = ift.full(pdom, 2.)
    o1 = ift.FFTOperator(f1.domain["a"])
    o2 = ift.FFTOperator(f1.domain["b"])
    op = (o1.ducktape("a").ducktape_left("a") +
          o2.ducktape("b").ducktape_left("b"))
    _, op2 = op.simplify_for_constant_input(f2)
    assert_allclose(op(f1)["a"].val, op2.force(f1)["a"].val)
    assert_allclose(op(f1)["b"].val, op2.force(f1)["b"].val)
    # FIXME Add test for ChainOperator._simplify_for_constant_input_nontrivial()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2022 Max-Planck-Society, Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from functools import reduce
from operator import add

import nifty8 as ift
import numpy as np
import pytest

from ..common import setup_function, teardown_function

pmp = pytest.mark.parametrize

doms = [ift.HPSpace(4), {"a": ift.UnstructuredDomain(2), "b": ift.HPSpace(4)}]

@pmp("with_names", [0, 1, 2])
@pmp("dom", doms)
def test_likelihood_single(with_names, dom):
    dom = ift.makeDomain(dom)
    data = ift.from_random(dom)
    icov = ift.from_random(dom).exp()
    lh = ift.GaussianEnergy(data=data, inverse_covariance=ift.makeOp(icov))
    if with_names == 1:
        lh.name = f"myname"
        assert lh.name == f"myname"
    if isinstance(dom, ift.DomainTuple):
        lh = lh.ducktape("b")
    if with_names == 2:
        lh.name = f"myname"
    if with_names in [1, 2]:
        assert lh.name == f"myname"
    print(lh)
    samples = ift.SampleList([ift.from_random(lh.domain) for _ in range(2)])
    print(ift.extra.minisanity(lh, samples))


@pmp("with_names", [0, 1, 2, 3])
def test_likelihood_sum(with_names):
    lhs = []
    for ii, dom in enumerate(doms):
        dom = ift.makeDomain(dom)
        data = ift.from_random(dom)
        icov = ift.from_random(dom).exp()
        lh = ift.GaussianEnergy(data=data, inverse_covariance=ift.makeOp(icov))
        if with_names == 1:
            lh.name = f"myname{ii}"
        if isinstance(dom, ift.DomainTuple):
            lh = lh.ducktape("b")
        if with_names == 2:
            lh.name = f"myname{ii}"
        if with_names == 3:
            lh.name = f"myname"
        lhs.append(lh)
    if with_names == 3:
        with pytest.raises(ValueError):
            lh = reduce(add, lhs)
        return
    lh = reduce(add, lhs)
    print(lh)
    samples = ift.SampleList([ift.from_random(lh.domain) for _ in range(2)])
    print(ift.extra.minisanity(lh, samples))



import pytest

pytest.importorskip("jax")

from functools import partial

import numpy as np
from numpy.testing import assert_allclose
from scipy import stats

import nifty8.re as jft

pmp = pytest.mark.parametrize


@pmp("a", (3., 1.5, 4.))
@pmp("scale", (2., 4.))
@pmp("loc", (2., 4., 0.))
@pmp("seed", (42, 43))
def test_invgamma_roundtrip(a, scale, loc, seed, step=1e-1):
    rng = np.random.default_rng(seed)

    n_samples = int(1e+4)
    n_rvs = rng.normal(loc=0., scale=2., size=(n_samples, ))
    n_rvs = n_rvs.clip(-5.2, 5.2)

    pr = jft.invgamma_prior(a, scale, loc=loc, step=step)
    ipr = jft.invgamma_invprior(a, scale, loc=loc, step=step)

    n_roundtrip = ipr(pr(n_rvs))
    assert_allclose(n_roundtrip, n_rvs, rtol=1e-4, atol=1e-3)


@pmp("mean", (2., 4.))
@pmp("std", (2., 4.))
@pmp("seed", (42, 43))
def test_lognormal_roundtrip(mean, std, seed):
    rng = np.random.default_rng(seed)

    n_samples = int(1e+4)
    n_rvs = rng.normal(loc=0., scale=2., size=(n_samples, ))

    pr = jft.lognormal_prior(mean, std)
    ipr = jft.lognormal_invprior(mean, std)

    n_roundtrip = ipr(pr(n_rvs))
    assert_allclose(n_roundtrip, n_rvs, rtol=1e-6, atol=1e-6)


@pmp(
    "stats_and_prior", (
        (stats.norm(), jft.normal_prior(mean=0, std=1)),
        (stats.laplace(), jft.laplace_prior(alpha=1)),
        (
            stats.lognorm(s=1),
            jft.lognormal_prior(None, None, _log_mean=0, _log_std=1)
        ),
        (stats.invgamma(a=2), jft.invgamma_prior(a=2, scale=1)),
        (stats.uniform(), jft.uniform_prior(a_min=0, a_max=1)),
        (stats.norm(), jft.NormalPrior(mean=0, std=1)),
        (stats.laplace(), jft.LaplacePrior(alpha=1)),
        (
            stats.lognorm(s=1),
            jft.LogNormalPrior(
                np.exp(0.5),
                np.exp(0.5) * np.sqrt(np.exp(1) - 1)
            )
        ),
        (stats.invgamma(a=2), jft.InvGammaPrior(a=2, scale=1)),
        (stats.uniform(), jft.UniformPrior(a_min=0, a_max=1)),
    )
)
def test_quantiles(stats_and_prior):
    stats_distr, prior_dist = stats_and_prior
    atol = 0.0
    rtol = 1e-9 if not stats_distr.dist.name == "invgamma" else 1e-5

    q = np.linspace(1e-6, 1 - 1e-6, num=100, endpoint=True)
    pp = stats.norm.ppf(q, loc=0., scale=1.)
    assert_allclose(prior_dist(pp), stats_distr.ppf(q), rtol=rtol, atol=atol)
